from .learners import (FairPredictor, inferred_attribute_builder, single_threshold, DataDict,
                       DeepFairPredictor, DeepDataDict)
from .utils import performance, group_metrics, conditional_group_metrics, dataset_loader

__all__ = (FairPredictor, inferred_attribute_builder, single_threshold, DataDict,
           performance, group_metrics, conditional_group_metrics, DeepFairPredictor, DeepDataDict, dataset_loader)



# Wrapper for scipy continuious metrics

from typing import Any
from sklearn.metrics import roc_auc_score, average_precision_score


class ScorerRequiresContPred:
    def __init__(self, scorer, name) -> None:
        self.scorer = scorer
        self.name = name
        self.greater_is_better = True
        self.cond_weights = None

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.scorer(*args, **kwds)


roc_auc = ScorerRequiresContPred(roc_auc_score, 'ROC AUC')
average_precision = ScorerRequiresContPred(average_precision_score, 'Average Precision')



import logging
import warnings
from subprocess import Popen, PIPE, check_call
import linkcheckmd as lc
import glob
# Owing to dependency issues with github actions this file is excluded from the tests.
# run pip install nbtest before running test_run_notebooks_without_errors


def test_check_style_codebase():
    logging.getLogger().setLevel(logging.INFO)
    logging.info("PEP8 Style check")
    flake8_proc = Popen(["flake8",   'src/oxonfair/',
                         "--count",
                         "--max-line-length", "150",],
                        stdout=PIPE)
    flake8_out = flake8_proc.communicate()[0]
    lines = flake8_out.splitlines()
    print(lines)
    count = int(lines[-1].decode())
    if count > 0:
        warnings.warn(f"{count} PEP8 warnings remaining")
    assert count < 10, "Too many PEP8 warnings found, improve code quality to pass test."


def test_check_style_tests():
    logging.getLogger().setLevel(logging.INFO)
    logging.info("PEP8 Style check")
    flake8_proc = Popen(["flake8",   'test',
                         "--count",
                         "--max-line-length", "150",],
                        stdout=PIPE)
    flake8_out = flake8_proc.communicate()[0]
    lines = flake8_out.splitlines()
    count = int(lines[-1].decode())
    if count > 0:
        warnings.warn(f"{count} PEP8 warnings remaining")
    assert count < 10, "Too many PEP8 warnings found, improve code quality to pass test."


def test_check_style_examples():
    logging.getLogger().setLevel(logging.INFO)
    logging.info("PEP8 Style check")
    flake8_proc = Popen(["flake8",   'examples',
                         "--count",
                         "--max-line-length", "150",],
                        stdout=PIPE)
    flake8_out = flake8_proc.communicate()[0]
    lines = flake8_out.splitlines()
    count = int(lines[-1].decode())
    if count > 0:
        warnings.warn(f"{count} PEP8 warnings remaining")
    assert count < 10, "Too many PEP8 warnings found, improve code quality to pass test."


def test_md_links():
    missing_links = lc.check_links('./', ext='.md')
    for link in missing_links:
        warnings.warn(link)
    assert missing_links == []


def test_run_notebooks_without_errors():
    "run pip install nbtest before running this test"
    check_call(['pytest', '--nbmake', '-n=auto', '--nbmake-timeout=500', 'examples'])


def test_compress_notebooks():
    # Now compress notebooks because running test makes them too large
    # This is not really a test, hijacking the test suite to build.
    from ipynbcompress import compress
    for file in glob.glob('./examples/*.ipynb'):
        compress(file)



"""Definitions of standard measures for fairness and performance"""

import numpy as np

from .scipy_metrics_cont_wrapper import roc_auc, average_precision, ScorerRequiresContPred  # noqa: F401
from .group_metric_classes import ( # pylint: disable=unused-import # noqa
    GroupMetric,
    AddGroupMetrics,
    Utility)  # noqa: F401
# N.B. BaseGroupMetric and Utility are needed for type declarations


def ge1(x):
    """Helper function.
    Return the elementwise maximum of x or 1.
    Used so that metrics of the form 0/0 are treated as 0 not NaN."""
    return np.maximum(1, x)


# Basic parity measures for fairness
count = GroupMetric(lambda TP, FP, FN, TN: TP + FP + FN + TN, "Number of Datapoints")
pos_data_count = GroupMetric(lambda TP, FP, FN, TN: TP + FN, "Positive Count")
neg_data_count = GroupMetric(lambda TP, FP, FN, TN: FP + TN, "Negative Count")
pos_data_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TP + FN) / (TP + FP + FN + TN), "Positive Label Rate"
)
neg_data_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TN + FP) / (TP + FP + FN + TN), "Negative Label Rate"
)
pos_pred_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TP + FP) / (TP + FP + FN + TN), "Positive Prediction Rate"
)
neg_pred_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TN + FN) / (TP + FP + FN + TN), "Negative Prediction Rate"
)

# Standard metrics see sidebar of https://en.wikipedia.org/wiki/Precision_and_recall
true_pos_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TP) / ge1(TP + FN), "True Positive Rate"
)
true_neg_rate = GroupMetric(
    lambda TP, FP, FN, TN: (TN) / ge1(FP + TN), "True Negative Rate"
)
false_pos_rate = GroupMetric(
    lambda TP, FP, FN, TN: (FP) / ge1(FP + TN), "False Positive Rate"
)
false_neg_rate = GroupMetric(
    lambda TP, FP, FN, TN: (FN) / ge1(TP + FN), "False Negative Rate"
)
pos_pred_val = GroupMetric(
    lambda TP, FP, FN, TN: (TP) / ge1(TP + FP), "Positive Predicted Value"
)
neg_pred_val = GroupMetric(
    lambda TP, FP, FN, TN: (TN) / ge1(TN + FN), "Negative Predicted Value"
)

# Existing binary metrics for autogluon
accuracy = GroupMetric(
    lambda TP, FP, FN, TN: (TP + TN) / (TP + FP + FN + TN), "Accuracy"
)
balanced_accuracy = GroupMetric(
    lambda TP, FP, FN, TN: (TP / ge1(TP + FN) + TN / ge1(TN + FP)) / 2,
    "Balanced Accuracy",
)
min_accuracy = GroupMetric(
    lambda TP, FP, FN, TN: np.minimum(TP / ge1(TP + FN), TN / ge1(TN + FP)),
    "Minimum-Label-Accuracy",
)  # common in min-max fairness literature
f1 = GroupMetric(lambda TP, FP, FN, TN: (2 * TP) / ge1(2 * TP + FP + FN), "F1 score")
precision = pos_pred_val.clone("Precision")
recall = true_pos_rate.clone("Recall")
mcc = GroupMetric(
    lambda TP, FP, FN, TN: (TP * TN - FP * FN)
    / ge1(np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))),
    "MCC",
)

bias_amplification = GroupMetric(lambda TP, FP, FN, TN: np.abs(FN - FP)/(TP + FP + FN + TN),
                                 'Absolute Bias Amplification', False)
# absolute metric based on the delta a_t term from directional bias amplification
# see: https://proceedings.mlr.press/v139/wang21t/wang21t.pdf

default_accuracy_metrics = {
    "accuracy": accuracy,
    "balanced_accuracy": balanced_accuracy,
    "f1": f1,
    "mcc": mcc,
}
additional_ag_metrics = {"precision": precision, "recall": recall, "roc_auc": roc_auc}
standard_metrics = {
    "true_pos_rate": true_pos_rate,
    "true_neg_rate": true_neg_rate,
    "false_pos_rate": false_pos_rate,
    "false_neg_rate": false_neg_rate,
    "pos_pred_val": pos_pred_val,
    "neg_pred_val": neg_pred_val,
}

ag_metrics = {**default_accuracy_metrics, **additional_ag_metrics}

count_metrics = {
    #   "count": count,
    "pos_data_count": pos_data_count,
    "neg_data_count": neg_data_count,
    "pos_data_rate": pos_data_rate,
    "pos_pred_rate": pos_pred_rate,
}
default_group_metrics = {**ag_metrics, **count_metrics}

extended_group_metrics = {
    **default_accuracy_metrics,
    **standard_metrics,
    **count_metrics,
}

# Postprocessing Clarify metrics
# https://mkai.org/learn-how-amazon-sagemaker-clarify-helps-detect-bias
class_imbalance = pos_data_rate.diff.clone("Class Imbalance")
demographic_parity = pos_pred_rate.diff.clone("Demographic Parity")
disparate_impact = pos_pred_rate.ratio.clone("Disparate Impact")
acceptance_rate = precision.clone("Acceptance Rate")
cond_accept = GroupMetric(
    lambda TP, FP, FN, TN: (TP + FN) / ge1(TP + FP), "Conditional Acceptance Rate"
)
cond_reject = GroupMetric(
    lambda TP, FP, FN, TN: (TN + FP) / ge1(TN + FN), "Conditional Rejectance Rate"
)
specificity = true_neg_rate.clone("Specificity")
rejection_rate = neg_pred_val.clone("Rejection Rate")
error_ratio = GroupMetric(lambda TP, FP, FN, TN: FP / ge1(FN), "Error Ratio")
treatment_equality = error_ratio.diff.clone("Treatment Equality")

gen_entropy = GroupMetric(
    lambda TP, FP, FN, TN: (
        (TP + FP + TN + FN) * (TP + FP * 4 + TN) / ge1(TP + FP * 2 + TN) ** 2 - 1
    )
    / 2,
    "Generalized Entropy",
    False,
)
clarify_metrics = {
    # "class_imbalance": class_imbalance,
    "demographic_parity": demographic_parity,
    "disparate_impact": disparate_impact,
    "cond_accept.diff": cond_accept.diff,
    "cond_reject.diff": cond_reject.diff,
    "accuracy.diff": accuracy.diff,
    "recall.diff": recall.diff,
    "acceptance_rate.diff": acceptance_rate.diff,
    "specificity.diff": specificity.diff,
    "rejection_rate.diff": rejection_rate.diff,
    "treatment_equality": treatment_equality,
    # gen_entropy": gen_entropy,
}


# Existing fairness definitions.
# Binary definitions from: https://fairware.cs.umass.edu/papers/Verma.pdf
# As all definitions just say 'these should be equal' we report the max difference in values
# as a measure of inequality.

statistical_parity = demographic_parity.clone("Statistical Parity")
predictive_parity = precision.diff.clone("Predictive Parity")
predictive_equality = false_neg_rate.diff.clone("Predictive Equality")
equal_opportunity = recall.diff.clone("Equal Opportunity")
equalized_odds = AddGroupMetrics(
    true_pos_rate.diff, true_neg_rate.diff, "Equalized Odds"
)
cond_use_accuracy = AddGroupMetrics(
    pos_pred_val.diff, neg_pred_val.diff, "Conditional Use Accuracy"
)
accuracy_parity = accuracy.diff.clone("Accuracy Parity")

verma_metrics = {
    "statistical_parity": statistical_parity,
    "predictive_parity": predictive_parity,
    "recall.diff": equal_opportunity,
    "miss_rate.diff": false_neg_rate.diff,
    "equalized_odds": equalized_odds,
    "cond_use_accuracy": cond_use_accuracy,
    # "predictive_equality": predictive_equality,
    "accuracy.diff": accuracy.diff,
    "treatment_equality": treatment_equality,
}

rate_metrics = {
    "pos_pred_rate": pos_pred_rate.diff,
    **{k: v.diff for k, v in standard_metrics.items()},
}

default_fairness_measures = verma_metrics

# Complex fairness definitions
# Here we define standard fairness measures that require more than per group difference/ratios

# We start with the unconditional forms of fairness from "why fairness can not be automated"
pos_pred_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: (TP+FP)/ge1(TTP+TFP)[:, np.newaxis],
                                  'Proportion of Positive Preditions', total_metric=True)
neg_pred_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: (TN+FN)/ge1(TTN+TFN)[:, np.newaxis],
                                  'Proportion of Negative Predictions', total_metric=True)
diff_pred_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: ((TP+FP)/ge1(TTP+TFP)[:, np.newaxis])
                                   - (TN+FN)/ge1(TTN+TFN)[:, np.newaxis],
                                   'Difference of Proportions of Predictions', total_metric=True)

abs_diff_pred_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: np.abs((TN+FN)/ge1(TTN+TFN)[:, np.newaxis]
                                                                                         - (TP+FP)/ge1(TTP+TFP)[:, np.newaxis]),
                                       'Proportion of Negative Predictions', total_metric=True)


pos_data_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: (TP+FN)/ge1(TTP+TFN)[:, np.newaxis],
                                  'Proportion of Positive Labels', total_metric=True)
neg_data_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: (TN+FP)/ge1(TTN+TFP)[:, np.newaxis],
                                  'Proportion of Negative Labels', total_metric=True)
diff_data_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: (TP+FN)/ge1(TTP+TFN)[:, np.newaxis]-(TN+FP)/ge1(TTN+TFP)[:, np.newaxis],
                                   'Difference of Proportion of Labels', total_metric=True)

abs_diff_data_proportion = GroupMetric(lambda TP, FP, FN, TN, TTP, TFP, TFN, TTN: np.abs((TP+FN)/ge1(TTP+TFN)[:, np.newaxis]
                                                                                         - (TN+FP)/ge1(TTN+TFP)[:, np.newaxis]),
                                       'Absolute Difference of Proportion of Labels', total_metric=True)

wachter_measures = {'pos_data_proportion': pos_data_proportion,
                    'neg_data_proportion': neg_data_proportion,
                    'diff_data_proportion': diff_data_proportion,
                    'pos_pred_proportion': pos_pred_proportion,
                    'neg_pred_proportion': neg_pred_proportion,
                    'diff_pred_proportion': diff_pred_proportion}

# directed bias amplification
# see: https://proceedings.mlr.press/v139/wang21t/wang21t.pdf
# warning do not use when enforcing fairness


def _y_a_t(TP, FP, FN, TN, TTP, TFN, TFP, TTN):
    "internal helper function that computes y_a_t from Wang et al"
    # p(A_a=1,T=1)
    # we remove one common normalisation term from PAT and PA as they cancel anyway
    PAT = (TP+FN)  # /(TTP+TFN+TFP+TTN)
    # p(A_a=1)p(T=1)
    PA = (TP+FP+FN+TN)  # /(TTP+TFN+TFP+TTN)
    PT = (TTP+TFN)/(TTP+TFN+TFP+TTN)
    out = 2*(PAT > PA*PT[:, np.newaxis]) - 1
    return out


directed_bias_amplification = GroupMetric(lambda TP, FP, FN, TN, TTP, TFN, TFP, TTN: _y_a_t(TP, FP, FN, TN, TTP, TFN, TFP, TTN) *
                                          (FN - FP)/(TP + FP + FN + TN),
                                          'Directed Bias Amplification', total_metric=True, greater_is_better=False)



"""The clases used to define group measures for fairness and performance"""

import abc
from abc import abstractmethod
import logging
import copy

from typing import Callable, Tuple
import numpy as np

logger = logging.getLogger(__name__)


class BaseGroupMetric:
    """Building block for GroupMetrics.
    It performs book keeping allowing group metrics to take as raw
    input either a single array containing t_pos,f_pos,f_neg,t_neg values broadcast over groups and
    many different thresholds, or singular vectors corresponding to y_true, y_pred, and groups.
    Also contains additional annotations: name, and greater_is_better
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        func: Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray],
        name: str,
        greater_is_better: bool, *,
        cond_weights=None,
        total_metric=False
    ) -> None:
        self.func: Callable = func
        self.name: str = name
        self.greater_is_better: bool = greater_is_better
        if cond_weights is None:
            self.cond_weights = None
        else:
            assert isinstance(
                cond_weights, ConditionalWeighting
            ), "cond_weights must be a Conditional Metric"
            self.cond_weights = cond_weights
        self.total_metric = total_metric

    @abstractmethod
    def __call__(self, *args: np.ndarray) -> np.ndarray:
        pass

    def build_array(
        self, args: Tuple[np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Helper Function for all child classes.
        Allows the overloading of GroupMetrics so they can be used both in the inner loop of
        efficient_compute.py and to return scores on raw data.
        parameters
        ----------
        args: a Tuple of numpy arrays.
            Either a Tuple containing a single (4 x entries x groups) or
                a Tuple containing 3 or 4 vectors of the same length corresponding to y_true, y_pred,
                groups and (optionally) weights
        returns
            if total_weights is False
            4  (entries x groups) sized arrays, where entries is 1 if args consisted of 3 or 4 vectors.
            if total_weights is True
            Two tupples consisting of
            4  (entries x groups) sized arrays, where entries is 1 if args consisted of 3 or 4 vectors.
            4 entiries size arrays
        """
        if len(args) == 1:
            assert args[0].shape[0] == 4, "Only one argument passed to group metric, but the first dimension is not 4."
            if self.total_metric is False:
                return args[0][3], args[0][2], args[0][1], args[0][0]
            else:
                aa = args[0].sum(2)
                return args[0][3], args[0][2], args[0][1], args[0][0], aa[3], aa[2], aa[1], aa[0]

        if len(args) == 2:
            assert args[0].shape[0] == 2, "Two arguments passed to group metric, but the first dimension is not 2."
            assert args[1].shape[0] == 2, "Two arguments passed to group metric, but the first dimension is not 2."
            if self.total_metric is False:
                return args[1][1], args[1][0], args[0][1], args[0][0]
            else:
                a1 = args[1].sum(1)
                a0 = args[0].sum(1)
                return args[1][1], args[1][0], args[0][1], args[0][0], a1[1], a1[0], a0[1], a0[0]
        assert len(args) <= 4, "Group metrics can only take up to 4 arrays as input"

        assert not (len(args) == 4 and self.cond_weights is None), ("Metric called with four inputs, indicating that we should "
                                                                    "condition but no conditioning function provided.")

        y_true: np.ndarray = args[0].astype(int)
        y_pred: np.ndarray = args[1].astype(int)
        groups: np.ndarray = args[2]
        if len(args) == 4:
            if groups.ndim == 2:
                weights = args[3]
            # if groups is a mask, weights must also be precomputed.
            else:
                weights = self.cond_weights(args[3], groups, y_true)
        else:
            weights = False

        assert y_true.size == y_pred.size == groups.shape[0], ("Inputs to group_metric are of different length. "
                                                               "Make sure that all variables are ordinal encoded and not one-hot.")
        t_pos = y_true * y_pred
        f_pos = (1 - y_true) * y_pred
        f_neg = y_true * (1 - y_pred)
        t_neg = (1 - y_true) * (1 - y_pred)
        if weights is not False:
            t_pos = t_pos * weights
            f_pos = f_pos * weights
            f_neg = f_neg * weights
            t_neg = t_neg * weights

        if groups.ndim == 2:
            out = np.zeros((4, 1, groups.shape[1]))
            out[0, 0] = t_pos.dot(groups)
            out[1, 0] = f_pos.dot(groups)
            out[2, 0] = f_neg.dot(groups)
            out[3, 0] = t_neg.dot(groups)
        else:
            unique = np.unique(groups)
            out = np.zeros((4, 1, unique.shape[0]))
            for i, group_name in enumerate(unique):
                mask = groups == group_name
                out[0, 0, i] = t_pos[mask].sum()
                out[1, 0, i] = f_pos[mask].sum()
                out[2, 0, i] = f_neg[mask].sum()
                out[3, 0, i] = t_neg[mask].sum()
        if self.total_metric is False:
            return out[0], out[1], out[2], out[3]
        else:
            aa = out.sum(2)
            return out[0], out[1], out[2], out[3], aa[0], aa[1], aa[2], aa[3]

    def clone(self, new_name: str, cond_weights=False):
        """Generates a copy of self with a new name, and (optionally) a new cond_weights
        Use cond_weights = False to use the existing cond_weights in the metric being cloned.
            cond_weights =  None to remove it
            cond_weights = ConditionalMetric in any other case"""
        out = copy.copy(self)
        out.name = new_name
        if cond_weights is not False:
            out.cond_weights = cond_weights
        return out


class PerGroup(BaseGroupMetric):
    "Helper class for reporting scores per group"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val


class GroupMax(BaseGroupMetric):
    "Helper class for reporting maximal score of any  group"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val.max(-1)


class GroupMin(BaseGroupMetric):
    "Helper class for reporting minimal score of any group"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val.min(-1)


class GroupMaxDiff(BaseGroupMetric):
    "Helper class for reporting maximal difference in score between any pair of groups"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val.max(-1) - val.min(-1)


class GroupDiff(BaseGroupMetric):
    "Helper class for reporting average difference in score between all pairs of groups"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        broadcast = val[:, np.newaxis, :] - val[:, :, np.newaxis]
        trunc = np.maximum(broadcast, 0)
        collate = trunc.sum(1).sum(1) / (val.shape[1] * (val.shape[1] - 1) / 2)
        return collate


class GroupRatio(BaseGroupMetric):
    "Helper class for reporting average score ratio  between any pair of groups"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        broadcast = val[:, np.newaxis, :] / val[:, :, np.newaxis]
        trunc = np.minimum(broadcast, 1.0/broadcast)
        trunc[~np.isfinite(trunc)] = 1
        idx = np.arange(trunc.shape[-1])
        trunc[:, idx, idx] = 0
        collate = trunc.sum(1).sum(1) / (val.shape[1] * (val.shape[1] - 1))
        return collate


class GroupMinimalRatio(BaseGroupMetric):
    "Helper class for reporting minimal score ratio  between any pair of groups"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val.min(-1) / np.maximum(1e-12, val.max(-1))


class Overall(BaseGroupMetric):
    "Helper class for reporting score over entire dataset"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        args = self.build_array(args)
        if self.total_metric:
            args = list(map(lambda x: x.sum(1), args[:4])) + list(args[4:])
        else:
            args = list(map(lambda x: x.sum(1), args))
        val = self.func(*args)
        return val


class GroupAverage(BaseGroupMetric):
    "Helper class for reporting scores averaged over groups"

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        array = self.build_array(args)
        val = self.func(*array)
        return val.mean(-1)


class GroupMetric(BaseGroupMetric):
    """Broadcastable metrics used by efficient compute.
    All methods either takes a single 3d numpy array as input or three vectors:
    y_true, y_pred, and groups
    The matrix passed to any function is assumed to be of size
    4 x entries x groups.
    The first entry of the first axis corresponds to the number of True Negatives,
    second False Negatives,
    third False Positives, and
    fourth True Positives.

    init parameters:
    func: a function that takes 4 numpy arrays corresponding to:
        True Positives, False Positives, False Negatives, and True Negatives as an input,
        and returns a numpy array of scores.
    name: a string description of the score.
    greater_is_better: a bool indicating if the score should be maximised or minimised.
    """

    def __init__(
        self,
        func: Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray],
        name: str,
        greater_is_better: bool = True,
        cond_weights=None,
        total_metric=False
    ) -> None:
        super().__init__(func, name, greater_is_better, cond_weights=cond_weights, total_metric=total_metric)
        self.max: GroupMax = GroupMax(
            func,
            "Maximal Group " + name,
            greater_is_better=False,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.min: GroupMin = GroupMin(
            func,
            "Minimal Group " + name,
            greater_is_better=True,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.overall: Overall = Overall(
            func,
            "Overall " + name,
            greater_is_better=greater_is_better,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.average: GroupAverage = GroupAverage(
            func,
            "Average Group " + name,
            greater_is_better=greater_is_better,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.diff: GroupDiff = GroupDiff(
            func,
            "Average Group Difference in " + name,
            greater_is_better=False,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.max_diff: GroupMaxDiff = GroupMaxDiff(
            func,
            "Maximal Group Difference in " + name,
            greater_is_better=False,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.ratio: GroupRatio = GroupRatio(
            func,
            "Average Group Ratio in " + name,
            greater_is_better=True,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.min_ratio: GroupMinimalRatio = GroupMinimalRatio(
            func,
            "Minimal Group Ratio in " + name,
            greater_is_better=True,
            cond_weights=cond_weights,
            total_metric=total_metric
        )
        self.per_group: PerGroup = PerGroup(
            func,
            "Per Group " + name,
            greater_is_better=greater_is_better,
            cond_weights=cond_weights,
            total_metric=total_metric
        )

    def clone(self, new_name, cond_weights=False):
        my_type = self.__class__
        if cond_weights is False:
            out = my_type(
                self.func, new_name, self.greater_is_better, 
                cond_weights=self.cond_weights, total_metric=self.total_metric
            )
        else:
            out = my_type(self.func, new_name, self.greater_is_better, cond_weights=cond_weights,
                          total_metric=self.total_metric)
        return out

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        return self.overall(*args)


class AddGroupMetrics(BaseGroupMetric):
    """Group Metric consisting of the weighted sum of two existing metrics
    parameters
    ----------
    metric1: a BaseGroupMetric
    metric2: a BaseGroupMetric
    name:    a string
    weight: (optional) a float between 0 and 1.
    returns
    -------
    a BaseGroupMetric that gives scores of the form:
        weight*metric1_response+(1-weight)*metric2_response"""

    def __init__(
        self,
        metric1: BaseGroupMetric,
        metric2: BaseGroupMetric,
        name: str,  # pylint: disable=super-init-not-called
        weight: float = 0.5,
    ) -> None:
        self.metric1: BaseGroupMetric = metric1
        self.metric2: BaseGroupMetric = metric2
        self.name = name
        self.cond_weights = None
        if metric1.greater_is_better != metric2.greater_is_better:
            logger.error(
                "Metric1 and metric2  must satisfy the condition. metric1.greater_is_better == metric2.greater_is_better "
            )
        if not 0 <= weight <= 1:
            logger.error("Weight must be between 0 and 1")
        self.weight: float = weight
        self.greater_is_better = metric1.greater_is_better

    def __call__(self, *args: np.ndarray) -> np.ndarray:
        return self.weight * self.metric1(*args) + (1 - self.weight) * self.metric2(
            *args
        )


class Utility(GroupMetric):
    """A group metric for encoding utility functions.
    See Fairness on the Ground: htt_poss://arxiv.org/pdf/2103.06172.pdf
    This is implemented as a group metric, so the standard fairness concerns i.e.
    difference in utility between groups, ratio of utility, minimum utility of any group
    are all supported.
    Parameters
    ----------
    utility: a sequence of length 4 corresponding to the cost of true positives,
             false positive, false negatives, and true negatives
    name: a string corresponding to the name of the utility function
    greater_is_better: a bool indicating if the utility should be maximised or minimized
    """

    def __init__(self, utility, name, greater_is_better=False):
        if len(utility) != 4:
            logger.error("Utility vector must be of length 4.")
        self.utility = utility
        super().__init__(self.cost, name, greater_is_better)

    def cost(self, t_pos, f_pos, f_neg, t_neg):
        "Method for computing the cost/utility"
        return (
            t_pos * self.utility[0]
            + f_pos * self.utility[1]
            + f_neg * self.utility[2]
            + t_neg * self.utility[3]
        ) / (t_pos + f_pos + f_neg + t_neg)


class ConditionalWeighting:
    """This is used to implement a range of conditional metrics analgous to those set out in Why fairness can not be automated and statistics
    (cite properly)
    The metrics are useful when you want to (for example) to ensure that individual schools should not discriminate in aggrigate
    against men and women, but different schools may have different acceptance rates, and have different rates of men
    and women applying there.
    """

    def __init__(self, per_group):
        self.per_group = per_group

    def __call__(
        self, conditioning_factor: np.array, groups: np.array, y_true: np.array
    ) -> np.array:
        assert conditioning_factor.shape == y_true.shape
        assert groups.shape[0] == y_true.shape[0]
        weights = np.zeros_like(y_true, dtype=float)
        uniq_f = np.unique(conditioning_factor)
        uniq_g = np.unique(groups)

        factor_masks = {}
        factor_weights = {}
        total_pos = y_true.sum()
        total_neg = y_true.shape[0] - total_pos
        for f in uniq_f:
            mask = conditioning_factor == f
            factor_masks[f] = mask
            factor_pos = y_true[mask].sum()
            factor_neg = mask.sum() - factor_pos
            factor_weights[f] = self.per_group(
                total_pos, total_neg, factor_pos, factor_neg
            )
        for g in uniq_g:
            mask = groups == g
            group_pos = y_true[mask].sum()
            group_neg = mask.sum() - group_pos
            for f in uniq_f:
                new_mask = mask * factor_masks[f]
                factor_sub_pos = y_true[new_mask].sum()
                factor_sub_neg = new_mask.sum() - factor_sub_pos
                weights[new_mask] = (
                    self.per_group(group_pos, group_neg, factor_sub_pos, factor_sub_neg)
                    / factor_weights[f]
                )
        return weights



"""Standard evaluation functionality
Contains 3 main functions:
1. evaluate which reports performance measures over the whole set.
2. evaluate_fairness which reports fairness measures over the whole set
3. evaluate_per_group which reports performance measures per group"""

import numpy as np
import pandas as pd
from . import group_metrics
from .group_metric_classes import BaseGroupMetric
from .scipy_metrics_cont_wrapper import ScorerRequiresContPred

AUTOGLUON_EXISTS = True
try:
    from autogluon.core.metrics import Scorer
except ModuleNotFoundError:
    AUTOGLUON_EXISTS = False


def evaluate(target, prediction, *,
             metrics=None, verbose=True, threshold="auto") -> pd.DataFrame:
    """Compute standard performance measures based on target values (i.e. y, or ground-truth)
        and predictions.
    parameters
    ----------
    target: the values a predictor should be making can be either values from {0,1} or {-1,1}
    prediction: the predictions made by the model
            predictions can be either logits or probabilities or discrete values, however,
            if discrete roc will not be informative.
    metrics: (optional) a dictionary where the keys are metric names and the elements are either
            scoreables or group metrics. If not provided report the standard metrics
            reported by autogluon on binary predictors
    verbose: (optional, bool default true) if the indexes in the returned dataframe should be long names (true),
            or the short names used as dictionary keys.
    returns
    -------
    a pandas dataset containing rows indexed by measure name
    """
    if metrics is None:
        metrics = group_metrics.ag_metrics

    threshold = find_threshold(threshold, prediction)
    groups = np.ones_like(prediction)

    return evaluate_fairness(
        target,
        prediction,
        groups,
        metrics=metrics,
        verbose=verbose,
        threshold=threshold)


def evaluate_fairness(target, prediction, groups, factor=None, *,
                      metrics=None, verbose=True, threshold="auto") -> pd.DataFrame:
    """Compute standard fairness metrics.

    parameters
    ----------
    parameters
    ----------
    target: the values a predictor should be making can be either values from {0,1} or {-1,1}
    prediction: the predictions made by the model
            predictions can be either logits or probabilities or discrete values, however,
            if discrete roc will not be informative.
    groups: numpy array containing discrete group labelling
    metrics: (optional) a dictionary where the keys are metric names and the elements are either
            scoreables or group metrics. If not provided report the standard metrics
            reported by AWS clarify
    verbose: (optional, bool) if the indexes in the returned dataframe should be long names (true),
            or the short names used as dictionary keys.
    returns
    -------
    a pandas dataset containing rows indexed by fairness measure name
    """
    threshold = find_threshold(threshold, prediction)
    if groups is None:
        groups = np.ones_like(target)

    if metrics is None:
        metrics = group_metrics.default_fairness_measures

    values = np.zeros(len(metrics))
    names = []
    for i, k in enumerate(metrics.keys()):
        if verbose is False:
            names.append(k)
        else:
            names.append(metrics[k].name)
        values[i] = dispatch_metric(
            metrics[k], target, prediction, groups, factor, threshold=threshold)

    return pd.DataFrame(values, index=names)


def evaluate_per_group(target, prediction, groups, factor=None, *,
                       metrics=None, threshold="auto", verbose=True):
    """Compute standard performance measures per group

    parameters
    ----------
    target: the values a predictor should be making can be either values from {0,1} or {-1,1}
    prediction: the predictions made by the model
            predictions can be either logits or probabilities or discrete values, however,
            if discrete roc will not be informative.
    groups: numpy array containing discrete group labelling
    metrics: (optional) a dictionary where the keys are metric names and the elements are either
            scoreables or group metrics. If not provided report the standard metrics
            reported by autogluon on binary predictors
    verbose: (optional, bool) if the indexes in the returned dataframe should be long names (true),
            or the short names used as dictionary keys.
    returns
    -------
    a pandas dataset containing rows indexed by fairness measure name
    """
    threshold = find_threshold(threshold, prediction)

    if metrics is None:
        metrics = group_metrics.default_group_metrics

    prediction = np.asarray(prediction)
    group_names = np.unique(groups)

    names = list(metrics.keys())

    overall_scores = list(
        map(lambda n: dispatch_metric(
                metrics[n], target, prediction, groups, factor, threshold=threshold),
            names)
        )
    scores = list(
        map(
            lambda n: dispatch_metric_per_group(
                metrics[n], target, prediction, groups, factor, threshold=threshold),
            names)
        )
    scores = np.stack(scores)
    overall_scores = np.stack(overall_scores)

    if verbose is False:
        pandas_names = names
    else:
        pandas_names = list(map(lambda n: metrics[n].name, names))
    gap = (scores.max(-1) - scores.min(-1)).reshape(-1, 1)
    collect = np.hstack((overall_scores.reshape(-1, 1), scores, gap))
    out = pd.DataFrame(
        collect.T,
        index=(["Overall"] + group_names.tolist() + ["Maximum difference"]),
        columns=pandas_names)
    if verbose:
        out.index.name = "Groups"
    else:
        out.index.name = "groups"
    return out


def dispatch_metric(metric, y_true, proba, groups, factor, *, threshold) -> float:
    """Helper function for making sure different types of Scorer and GroupMetrics get the right data

    Parameters
    ----------
    metric: a BaseGroupMetric or Scorable
    y_true: a numpy array indicating positive or negative labels
    proba: a 2xdatapoints numpy or pandas array
    groups: a numpy array indicating group membership.

    Returns
    -------
     a numpy array containing the score provided by metrics
    """

    threshold = find_threshold(threshold, proba)
    # y_true might take values other than +1,-1
    y_true = y_true > 0

    proba = np.asarray(proba)
    try:
        if isinstance(metric, BaseGroupMetric):
            if metric.cond_weights is None:
                return metric(y_true, proba > threshold, groups)[0]
            return metric(y_true, proba > threshold, groups, factor)[0]

        if isinstance(metric, ScorerRequiresContPred) or (
                AUTOGLUON_EXISTS
                and isinstance(metric, Scorer)
                and (metric.needs_pred is False)):
            return metric(y_true, proba)
        return metric(y_true, proba > threshold)
    except ValueError:
        return np.nan


def dispatch_metric_per_group(
        metric, y_true: np.ndarray, proba: np.ndarray, groups: np.ndarray, factor: np.ndarray, *,
        threshold="auto") -> np.ndarray:
    """Helper function for making sure different types of Scorer and GroupMetrics get the right data
    parameters
    ----------
    metric: a GroupMetric or Scorable
    y_true: a binary numpy array indicating positive or negative labels
    proba: a 2xdatapoints numpy or pandas array
    groups: a numpy array indicating group membership.

    returns
    -------
    a numpy array containing the per group score provided by metrics"""
    threshold = find_threshold(threshold, proba)

    # y_true might take values other than +1,-1
    y_true = y_true > 0

    if isinstance(metric, group_metrics.GroupMetric):
        if metric.cond_weights is None:
            return metric.per_group(y_true, proba > threshold, groups)[0]
        return metric.per_group(y_true, proba > threshold, groups, factor)[0]
    unique = np.unique(groups)
    out = np.empty_like(unique, dtype=float)
    if isinstance(metric, ScorerRequiresContPred) or (
            AUTOGLUON_EXISTS
            and isinstance(metric, Scorer)
            and (metric.needs_pred is False)):
        for i, grp in enumerate(unique):
            mask = grp == groups
            try:
                out[i] = metric(y_true[mask], proba[mask])
            except ValueError:
                out[i] = np.nan
    else:
        out = metric(y_true, proba > threshold, groups)

    return out


def find_threshold(threshold, predictions):
    "heleper fn to automatically decide if we are dealing with logits or sigmoids"
    if threshold == "auto":
        if predictions.min() < 0 or predictions.max() > 1.0:
            return 0
        return 0.5
    return threshold



import numpy as np
import pandas as pd
from oxonfair import DataDict
from sklearn.preprocessing import LabelEncoder


class resample:
    "Induce unfairness by artifically altering the base rate and dropping positive samples for one group"
    def __init__(self, drop_group, drop_label, proportion) -> None:
        self.drop_group = drop_group
        self.drop_label = drop_label
        self.proportion = proportion

    def __call__(self, groups, labels):
        mask = np.random.random_sample(size=labels.shape) < self.proportion
        mask += (groups != self.drop_group)
        mask += (labels != self.drop_label)
        return mask > 0


def uniform_partition(target, groups, train_prop, test_prop, seed=None):
    "sample datasets uniformly. Datasets are assigned 0 for train, 1 for test and 2 for validation"
    target = np.asarray(target).reshape(-1)
    groups = np.asarray(groups).reshape(-1)
    assert target.shape == groups.shape
    out = np.empty_like(target, dtype=int)
    if seed is not None:
        np.random.seed(seed)
    for g in np.unique(groups):
        for t in np.unique(target):
            mask = (target == t) * (groups == g)
            size = mask.sum()
            dummy = np.empty(size)
            train_size = round(size*train_prop)
            test_size = train_size+round(size*test_prop)
            dummy[:train_size] = 0
            dummy[train_size:test_size] = 1
            dummy[test_size:] = 2
            np.random.shuffle(dummy)
            out[mask] = dummy
    return out


class partition:
    "Generic dataset loader and partitioner"
    def __init__(self, get_data, default_groups=None, resample=None) -> None:
        self.get_data = get_data
        self.resample = resample
        self.default_groups = default_groups

    def __call__(self,  groups=None, train_proportion=0.5, test_proportion=0.25, *,
                 seed=None, discard_groups=False, replace_groups=False,
                 encoding='ordinal', resample=None, seperate_groups=False):
        """Generic code for controling datapartitioning.
        groups: a specification of the column containing group information that can be understood by pandas
        train_proportion: number between 0 and 1 expressing the proportion of the dataset used for training
        test_proportion: as above but for test
        seed: random seed used to make results deterministic
        discard_groups: if True drop groups from the data
        replace_groups: A dict used to merge small groups e.g.
                    {'Hispanic':'Other', 'Native American':'Other', 'Asian':'Other'}
        encoding: if 'ordinal' or 'onehot' encode data accordingly. If None don't encode.
        resample: override existing resampling. This should be a Resample class.
        seperate_groups: default False. This indicates if groups should be stored as a seperable human
                        readable array or kept as a string.
                         Should be False if you don't want to explicitly pass groups to predict.
        """
        assert groups is not None or self.default_groups is not None
        if groups is None:
            groups = self.default_groups
        if resample is None:
            resample = self.resample

        total_data, target, positive_target = self.get_data()

        if callable(groups):
            groups = groups(total_data)

        if isinstance(target, str):
            t_name = target
            target = total_data[target]
            total_data = total_data.drop(t_name, axis=1)
        target = np.asarray(target).reshape(-1)
        if positive_target:
            if callable(positive_target):
                target = positive_target(target)
            else:
                target = target == positive_target
        assert all(0 <= target), 'target must be binary, or provide positive_target value'
        assert all(target <= 1), 'target must be binary, or provide positive_target value'
        assert 0 < target.mean() < 1, 'Something is wrong with the dataset. Every target value is the same.'

        assert 0 < np.asarray(target).mean() < 1

        if isinstance(groups, str):
            g_name = groups
            if replace_groups:
                total_data[groups] = total_data[groups].replace(replace_groups)

            if seperate_groups:
                groups = total_data[groups]

            if discard_groups:
                groups = total_data[g_name]
                total_data = total_data.drop(g_name, axis=1)

        else:
            if replace_groups:
                groups = groups.replace(replace_groups)

        if resample:
            mask = resample(groups, target)
            total_data = total_data[mask]
            if discard_groups or replace_groups:
                groups = groups[mask]
            target = target[mask]

        total_data.reset_index(drop=True)
        if not isinstance(groups, str):
            groups.reset_index(drop=True)

        if encoding == 'onehot':
            total_data = total_data.get_dummies()
        elif encoding == 'ordinal':
            total_data = total_data.apply(LabelEncoder().fit_transform)
        elif encoding is not None:
            assert encoding is not None, "encoding must be 'onehot', 'ordinal', or None"

        if not isinstance(groups, str):
            part = uniform_partition(target, groups, train_prop=train_proportion,
                                     test_prop=test_proportion, seed=seed)
            train_groups = groups.iloc[part == 0]
            val_groups = groups.iloc[part == 2]
            test_groups = groups.iloc[part == 1]
        else:
            part = uniform_partition(target, total_data[groups], train_prop=train_proportion,
                                     test_prop=test_proportion, seed=seed)
            train_groups = groups
            val_groups = groups
            test_groups = groups

        train = total_data.iloc[part == 0]
        train_y = target[part == 0]

        val = total_data.iloc[part == 2]
        val_y = target[part == 2]

        test = total_data.iloc[part == 1]
        test_y = target[part == 1]

        train_dict = DataDict(train_y, train, train_groups)
        val_dict = DataDict(val_y, val, val_groups)
        test_dict = DataDict(test_y, test, test_groups)
        return train_dict, val_dict, test_dict


def adult_raw():
    train_data = pd.read_csv("https://autogluon.s3.amazonaws.com/datasets/Inc/train.csv")
    test_data = pd.read_csv("https://autogluon.s3.amazonaws.com/datasets/Inc/test.csv")
    return pd.concat([train_data, test_data]), 'class', ' >50K'


def compas_raw():
    all_data = pd.read_csv('https://github.com/propublica/compas-analysis/raw/master/compas-scores-two-years.csv')
    condensed_data = all_data[['sex', 'race', 'age', 'juv_fel_count', 'juv_misd_count', 'juv_other_count',
                               'priors_count', 'age_cat', 'c_charge_degree', 'two_year_recid']].copy()
    return condensed_data, 'two_year_recid', None


def compas_audit_raw():
    all_data = pd.read_csv('https://github.com/propublica/compas-analysis/raw/master/compas-scores-two-years.csv')
    condensed_data = all_data[['sex', 'race', 'age', 'juv_fel_count', 'juv_misd_count', 'juv_other_count',
                               'priors_count', 'age_cat', 'c_charge_degree', 'decile_score.1',
                               'v_score_text', 'two_year_recid']].copy()
    return condensed_data, 'two_year_recid', None


def german_col_names(X):
    X.columns = ['status', 'duration', 'history', 'purpose', 'amount', 'savings/bonds', 'employment since',
                 'percent disposable', 'marital status', 'debtors', 'residence since', 'property', 'age',
                 'other plans', 'housing', 'existing credits', 'job', 'people liable', 'telephone', 'foreign worker',]

    X = X.replace(german_dict)
    return X


def taiwan_col_names(X):
    X.columns = ['Limit balance', 'sex', 'education', 'marriage', 'age', 'pay1', 'pay2', 'pay3', 'pay4', 'pay5',
                 'pay6', 'bill amount1', 'bill amount2', 'bill amount3', 'bill amount4', 'bill amount5', 'bill amount6',
                 'pay amount1', 'pay amount2', 'pay amount3', 'pay amount4', 'pay amount5', 'pay amount6']
    return X


def german_sex(X):
    return pd.DataFrame(list(map(lambda x: x.startswith('female'), X['marital status'])))


class UCI_raw:
    def __init__(self, index, pos_y_val=None, fix_X=None, fix_y=None) -> None:
        """index is the UCI index used to specify the dataset
        fixX is any function that must be called on X to clean it.
        fixY is a function that must be called on y to clean it
        """
        self.index = index
        self.fix_X = fix_X
        self.fix_y = fix_y
        self.pos_y_val = pos_y_val

    def __call__(self):
        from ucimlrepo import fetch_ucirepo
        data = fetch_ucirepo(id=self.index)

        # data (as pandas dataframes)
        X = data.data.features
        if self.fix_X is not None:
            X = self.fix_X(X)
        y = data.data.targets
        if self.fix_y is not None:
            y = self.fix_y(y)
        return X, y, self.pos_y_val


def replace_nan(X):
    X[X.isnull()] = -1
    return X


diabetes_raw = UCI_raw(891)
support2_raw = UCI_raw(880, fix_y=lambda y: y['death'])
german_raw = UCI_raw(144, pos_y_val=2, fix_X=german_col_names,)
taiwan_default_raw = UCI_raw(350, fix_X=taiwan_col_names)
bank_marketing_raw = UCI_raw(222, pos_y_val='yes')
student_raw = UCI_raw(856, fix_y=lambda y: y >= 5)
myocardial_infarction_raw = UCI_raw(579, fix_X=replace_nan, fix_y=lambda y: y['LET_IS'] > 0)

adult = partition(adult_raw, 'sex')
compas = partition(compas_raw, 'race')
compas_audit = partition(compas_audit_raw, 'race')
diabetes = partition(diabetes_raw, 'Sex')
support2 = partition(support2_raw, 'sex')
german = partition(german_raw, german_sex)
taiwan_default = partition(taiwan_default_raw, 'sex')
bank_marketing = partition(bank_marketing_raw, 'marital')
student = partition(student_raw, 'Sex')
myocardial_infarction = partition(myocardial_infarction_raw, 'SEX')

german_dict = {
        'A11': '< 0 DM',
        'A12': '< 200 DM',
        'A13': '>= 200 DM / salary assignments for at least 1 year',
        'A14': 'no checking account',
        'A30': 'no credits taken/all credits paid back duly',
        'A31': 'all credits at this bank paid back duly',
        'A32': 'existing credits paid back duly till now',
        'A33': 'delay in paying off in the past',
        'A34': 'critical account/ other credits existing (not at this bank)',
        'A40': 'car (new)',
        'A41': 'car (used)',
        'A42': 'furniture/equipment',
        'A43': 'radio/television',
        'A44': 'domestic appliances',
        'A45': 'repairs',
        'A46': 'education',
        'A47': '(vacation - does not exist?)',
        'A48': 'retraining',
        'A49': 'business',
        'A410': 'others',
        'A61': '         ... <  100 DM',
        'A62': '  100 <= ... <  500 DM',
        'A63': '  500 <= ... < 1000 DM',
        'A64': '         .. >= 1000 DM',
        'A65': '  unknown/ no savings account',
        'A71': 'unemployed',
        'A72': '      ... < 1 year',
        'A73': '1  <= ... < 4 years',
        'A74': '4  <= ... < 7 years',
        'A75': '      .. >= 7 years',
        'A91': 'male   :divorced/separated',
        'A92': 'female :divorced/separated/married',
        'A93': 'male   :single',
        'A94': 'male   :married/widowed',
        'A95': 'female :single',
        'A101': 'none',
        'A102': 'co-applicant',
        'A103': 'guarantor',
        'A121': 'real estate',
        'A122': 'if not A121:building society savings agreement/life insurance',
        'A123': 'if not A121/A122:car or other, not in attribute 6',
        'A124': 'unknown / no property',
        'A141': 'bank',
        'A142': 'stores',
        'A143': 'none',
        'A151': 'rent',
        'A152': 'own',
        'A153': 'for free',
        'A171': 'unemployed/ unskilled  - non-resident',
        'A172': 'unskilled - resident',
        'A173': 'skilled employee / official',
        'A174': 'management/ self-employed/highly qualified employee/ officer',
        'A191': 'none',
        'A192': 'yes, registered under the customers name',
        'A201': 'yes',
        'A202': 'no'
}



"""Implements efficient methods for fast computation of binary metrics"""
import logging
from typing import Callable, Tuple,  List, Sequence, Union
import numpy as np
from sklearn.preprocessing import OrdinalEncoder
from ..utils.group_metric_classes import BaseGroupMetric
logger = logging.getLogger(__name__)


def compute_metric(
        metric: Callable,
        y_true: np.ndarray,
        proba: np.ndarray,
        groups: np.ndarray,
        group_prediction: np.ndarray,
        weights: np.ndarray) -> np.ndarray:
    """takes probability scores, and offsets them according to weights[group_prediction].
        then selects the max and computes a fairness metric

    Parameters
    ----------
    metric: a BaseGroupMetric
    y_true: a numpy array containing the target labels
    proba: a numpy array containing the soft classifier responses.
    groups: a numpy array containing group assignment
    threshold assignment: a numpy array containing group predictions, when groups are infered
        this differs from groups
    weights: a numpy array containing the set of per group offsets
    Returns
    -------
    a numpy array of scores for each choice of weight
    """

    score = np.zeros((weights.shape[-1]))
    y_true = np.asarray(y_true)
    group_prediction = group_prediction.astype(int)
    for i in range(weights.shape[-1]):
        proba_update = proba.copy()
        proba_update[:, 1] += weights[group_prediction, i]
        pred = proba_update.argmax(-1)
        score[i] = metric(y_true, pred, groups)[0]
    return score


def keep_front(solutions: np.ndarray, initial_weights: np.ndarray, directions: np.ndarray,
               additional_constraints: Sequence,
               *, tol=1e-12) -> Tuple[np.ndarray, np.ndarray]:
    """Modified from
        https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python
        Returns Pareto efficient row subset of solutions and its associated weights
        Direction if a vector that governs if the frontier should maximize or minimize each
        direction.
        Where an element of direction is positive frontier maximizes, negative, it mimizes.
        parameters
        ----------
        solutions: a numpy array of values that are evaluated to find the frontier
        initial_weights: a numpy array of corresponding weights
        directions: a binary vector containing [+1,-1] indicating if greater or lower solutions are
            prefered
        tol: a float indicating if points that are almost dominated (i.e. they are within tol of
            another point in the frontier)  should be dropped.
            This is used to eliminate ties, and to discard most of the constant classifiers.
        additional constrains: vector of floats of size frontier width - 2
            These are hard constraints any point will be discarded if
            solution[i+2]*direction<additional_constraints[i]*direction .
        returns
        -------
        a pair of numpy arrays.
            1. reduced set of solutions associated with the Pareto front
            2. reduced set of weights associated with the Pareto front
    """

    front = solutions.T.copy()
    weights = initial_weights.T.copy()
    weights = weights.reshape(weights.shape[0], -1)  # handle both cases.
    front *= directions[:front.shape[1]]
    # drop all Nans/Inf
    mask = np.isfinite(front).any(1)
    front = front[mask]
    weights = weights[mask]
    # drop all points violating additional constraints.
    for i, val in enumerate(additional_constraints):
        mask = front[:, 2+i] >= val*directions[2+i]
        front = front[mask]
        weights = weights[mask]

    # drop all points worse than the extrema of the front
    # NB we have ties so pick the best extrema
    # This matters for replicability rather than performance
    best0 = front[:, 0] == front[:, 0].max()
    best1 = front[:, 1] == front[:, 1].max()
    ext1 = front[best0, 1].max()
    ext0 = front[best1, 0].max()
    mask = np.greater_equal(front[:, 1], ext1)
    mask *= np.greater_equal(front[:, 0], ext0)
    front = front[mask]
    weights = weights[mask]
    # sort points by decreasing sum of coordinates
    # Add 10**-8 * magnitude of w so that in the event of a near tie, pick points close to
    # the mean first
    mean = weights.mean(0)
    modifier = -(10**-8) * np.abs(weights - mean).sum(1)
    # code silently breaks if :2 is removed from front,
    order = (front[:, :2].sum(1) + modifier).argsort(kind='stable')[::-1]
    front = front[order]
    weights = weights[order]
    # initialize a boolean mask for currently undominated points
    undominated = np.ones(front.shape[0], dtype=bool)

    for i in range(front.shape[0]):
        size = front.shape[0]
        # process each point in turn
        if i >= size:
            break
        # find all points not dominated by i
        # since points are sorted by coordinate sum
        # i cannot dominate any points in 1,...,i-1
        undominated[i] = True  # Bug fix missing from online version
        undominated[i + 1:size] = (front[i + 1:size, :2] >= front[i, :2] + tol).any(1)
        front = front[undominated[:size]]
        weights = weights[undominated[:size]]

    weights = weights.T
    front *= directions[:front.shape[1]]
    # front = front[:, :2]
    front = front.T
    order = (front[0]).argsort()
    front = front[:, order]
    weights = weights[:, order]
    if initial_weights.ndim == 3:
        weights = weights.reshape(initial_weights.shape[1], initial_weights.shape[0], -1)
        weights = weights.transpose(1, 0, 2)
    return front, weights


def build_grid_inner(accum_count, mesh, groups):
    """Sample from accum_count using mesh"""
    acc = accum_count[0][mesh[0]]
    for i in range(1, len(accum_count)):
        acc = acc + accum_count[i][mesh[i]]  # variable matrix size mean += doesn't work
    assert acc.shape[-2:] == (4, groups)
    acc = acc.reshape(-1, acc.shape[-2], groups)
    acc = acc.transpose(1, 0, 2)

    return acc


def build_grid(accum_count: np.ndarray, bottom, top, metrics: Tuple[Callable],
               *, steps=25) -> Tuple[np.ndarray, List[np.ndarray], np.ndarray]:
    """Part of efficient grid search.
    This uses the fact that metrics can be computed efficiently as a function of TP,FP,FN and TN.
    By sorting the data per assigned group  we can efficiently compute these four values by looking
    at the cumlative sum of positive and negative labelled data (provided by ordered encode).
    This brings any subsequent computation of metrics down to O(1) in the dataset size.
    Parameters
    ----------
    accum_count:
    bottom: a single number or per group numpy array indicating where the grid should start
    top: a single number or per group numpy array indicating where the grid should stop
    metrics: an iterable of BaseGroupMetrics
    steps: (optional) The number of divisions per group
    returns
    -------
    a tupple of three numpy arrays:
        1. the scores of metrics computed for each choice of threshold
        2. the indicies corresponding to thresholds
        3. the step offset used.
    """
    groups = accum_count[0].shape[-1]
    step = [(t - b) / steps for t, b in zip(top, bottom)]
    mesh_indices = [np.unique(np.floor(np.arange(np.floor(b), np.ceil(t + 1),
                                                 max(1, s))).astype(int))
                    for b, t, s in zip(bottom, top, step)]
    mesh = np.meshgrid(*mesh_indices, sparse=True)

    grid = build_grid_inner(accum_count, mesh, groups)
    collect = [metric(grid) for metric in metrics]
    score = np.stack(collect, 0)
    return score, mesh_indices, np.maximum(1, np.asarray(step))


def build_grid2(accum_counts: Tuple[np.ndarray], bottom, top, metrics: Tuple[Callable],
                *, steps=25) -> Tuple[np.ndarray, List[np.ndarray], np.ndarray]:
    """Part of efficient grid search.
    This uses the fact that metrics can be computed efficiently as a function of TP,FP,FN and TN.
    By sorting the data per assigned group  we can efficiently compute these four values by looking
    at the cumlative sum of positive and negative labelled data (provided by ordered encode).
    This brings any subsequent computation of metrics down to O(1) in the dataset size.
    Parameters
    ----------
    accum_count:
    bottom: a single number or per group numpy array indicating where the grid should start
    top: a single number or per group numpy array indicating where the grid should stop
    metrics: an iterable of BaseGroupMetrics
    steps: (optional) The number of divisions per group
    returns
    -------
    a tupple of three numpy arrays:
        1. the scores of metrics computed for each choice of threshold
        2. the indicies corresponding to thresholds
        3. the step offset used.
    """
    groups = accum_counts[0][0].shape[-1]

    step = [(t - b) / steps for t, b in zip(top, bottom)]

    mesh_indices = [np.unique(np.floor(np.arange(np.floor(b), np.ceil(t + 1),
                                                 max(1, s))).astype(int))
                    for b, t, s in zip(bottom, top, step)]
    mesh = np.meshgrid(*mesh_indices, sparse=True)
    collect = [metric(build_grid_inner(acc, mesh, groups)) for acc, metric in zip(accum_counts, metrics)]

    score = np.stack(collect, 0)
    return score, mesh_indices, np.maximum(1, np.asarray(step))


def condense(thresholds: np.ndarray, labels: np.ndarray, lmax: int, groups: np.ndarray,
             gmax: int) -> Tuple[np.ndarray, np.ndarray]:
    """Take an array of float thresholds and non-negative integer labels, groups and
    return a sorted List of unique thresholds and the counts for each unique count of
    threshold, label, group
    parameters
    ----------
    thresholds: a numpy array of initial thresholds to reduce.
    labels: a numpy array of initial labels to count when reducing.
    lmax: the maximum label ever used.
    groups: a numpy array of initial groups to count when reducing.
    gmax: the maximum group ever used
    returns
    -------
    1. a sorted numpy array of unique thresholds
    2. corresponding counts
    """
    assert thresholds.shape == labels.shape == groups.shape
    groups = groups.astype(int)
    labels = labels.astype(int)
    unique_thresh, index = np.unique(thresholds, return_inverse=True)
    out = np.zeros((unique_thresh.shape[0], lmax, gmax), dtype=float)
    # float is noticably faster for higher numbers of groups.
    np.add.at(out, (index, labels, groups), 1)

    # assert out.sum() == labels.shape[0]

    return unique_thresh[::-1], out[::-1]


def condense_weights(thresholds: np.ndarray, labels: np.ndarray, lmax: int, groups: np.ndarray,
                     gmax: int, *, weights) -> Tuple[np.ndarray, np.ndarray]:
    """Take an array of float thresholds and non-negative integer labels, groups and
    return a sorted List of unique thresholds and the counts for each unique count of
    threshold, label, group
    parameters
    ----------
    thresholds: a numpy array of initial thresholds to reduce.
    labels: a numpy array of initial labels to count when reducing.
    lmax: the maximum label ever used.
    groups: a numpy array of initial groups to count when reducing.
    gmax: the maximum group ever used
    weights1 and weights2
    returns
    -------
    1. a sorted numpy array of unique thresholds
    2. two corresponding counts
    """
    assert thresholds.shape == labels.shape == groups.shape
    for w in weights:
        if isinstance(w, np.ndarray):
            assert w.shape == thresholds.shape
    groups = groups.astype(int)
    labels = labels.astype(int)
    unique_thresh, index = np.unique(thresholds, return_inverse=True)
    out = np.zeros((len(weights), unique_thresh.shape[0], lmax, gmax))
    for i in range(len(weights)):
        np.add.at(out[i], (index, labels, groups), weights[i])
        out[i] = out[i][::-1]

    return unique_thresh[::-1], out


def test_cum_sum(accum_count, groups):
    "Check expected properties of accum_count hold"
    # N.B. all values are int, and float approximation is not a concern
    for group in range(groups):
        assert (np.abs(accum_count[group].sum(1) - accum_count[group][0].sum(0)).sum()) == 0
        # Total sum must be the same
        assert (np.abs(accum_count[group][:, 0] + accum_count[group][:, 2]
                       - accum_count[group][0][0] - accum_count[group][0][2]).sum()) == 0
        # TP+FN must be the same
        assert (np.abs(accum_count[group][:, 1] + accum_count[group][:, 3]
                       - accum_count[group][0][1] - accum_count[group][0][3]).sum()) == 0
        # FP+TN must be the same


def cumsum_zero(array: np.ndarray):
    "compute a cumalitive sum starting with zero (i.e. the sum upto the first element)"
    zero = np.zeros((1,) + array.shape[1:], dtype=int)
    out = np.concatenate((zero, np.cumsum(array, 0)), 0)
    return out


def cumsum_zero_and_reverse(array: np.ndarray):
    """compute a cumalitive sum starting with zero (i.e. the sum upto the first element).
    Then reverse this by subtracting it from it's final elemenets and concatinate the two arrays"""
    cum_array = cumsum_zero(array)
    reverse_array = cum_array[-1].copy() - cum_array
    return np.concatenate((cum_array, reverse_array), 1)


def grid_search_no_weights(ordered_encode, ass_size, score,
                           metrics, steps, directions, additional_constraints):
    """Internal helper for grid search.
    The weighted pathway requires x2 memory and computation so instead of compressing the cases
    and computing unweighted as weighted with weights 1, we preserve the old pathway."""

    accum_count = [cumsum_zero_and_reverse(o) for o in ordered_encode]
    # The above is the important code
    # accum_count is a list of size groups where each element is an array consisting of the number
    # of true positives, false positives, false negatives and false positives if a threshold is set
    # at a particular value. It is of size (4, groups) because the group assignment may come at test
    # time from an inaccurate classifier

    # test_cum_sum(accum_count, ass_size)
    # now for the computational bottleneck
    bottom = np.zeros(ass_size)
    top = np.asarray([s.shape[0] for s in ordered_encode])
    score, mesh_indices, step = build_grid(accum_count, bottom, top, metrics, steps=steps)

    indicies = np.asarray(np.meshgrid(*mesh_indices, sparse=False)).reshape(ass_size, -1)

    front, index = keep_front(score, indicies, directions, additional_constraints)
    if index.shape[1] > 4:  # drop the absolute extrema
        tindex = index[:, 1:-1]
    else:
        tindex = index
    bottom = np.floor(np.maximum(step / 2, tindex.min(1) - step))
    top = np.ceil(np.minimum(top, tindex.max(1) + step))
    score, mesh_indices, _ = build_grid(accum_count, bottom, top, metrics, steps=steps)

    indicies = np.asarray(np.meshgrid(*mesh_indices, sparse=False)).reshape(ass_size, -1)
    return score, indicies, front, index


def grid_search_weights(ordered_encode, ordered_encode2, groups, score,
                        metrics, steps, directions, additional_constraints):
    """Internal helper for grid search.
    The weighted pathway requires x2 memory and computation so instead of compressing the cases
    and computing unweighted as weighted with weights 1, we preserve the old pathway.
    This is the new pathway for weights"""

    accum_count1 = [cumsum_zero_and_reverse(o) for o in ordered_encode]
    accum_count2 = [cumsum_zero_and_reverse(o) for o in ordered_encode2]

    # The above is the important code
    # accum_count is a list of size groups where each element is an array consisting of the number
    # of true positives, false positives, false negatives and false positives if a threshold is set
    # at a particular value. It is of size (4, groups) because the group assignment may come at test
    # time from an inaccurate classifier

    # now for the computational bottleneck
    bottom = np.zeros(groups)
    top = np.asarray([s.shape[0] for s in ordered_encode])
    score, mesh_indices, step = build_grid2((accum_count1, accum_count2), bottom, top, metrics,
                                            steps=steps)

    indicies = np.asarray(np.meshgrid(*mesh_indices, sparse=False)).reshape(groups, -1)

    front, index = keep_front(score, indicies, directions, additional_constraints)
    if index.shape[1] > 4:  # drop the absolute extrema
        tindex = index[:, 1:-1]
    else:
        tindex = index
    bottom = np.floor(np.maximum(step / 2, tindex.min(1) - step))
    top = np.ceil(np.minimum(top, tindex.max(1) + step))
    score, mesh_indices, _ = build_grid2((accum_count1, accum_count2), bottom, top, metrics,
                                         steps=steps)

    indicies = np.asarray(np.meshgrid(*mesh_indices, sparse=False)).reshape(groups, -1)
    return score, indicies, front, index


def grid_search(y_true: np.ndarray, proba: np.ndarray, metrics: Tuple[BaseGroupMetric],
                hard_assignment: np.ndarray, true_groups: np.ndarray, *, directions=(+1, +1),
                group_response=False, steps=25, factor=None,
                additional_constraints=()) -> Tuple[np.ndarray, np.ndarray]:
    """Efficient grid search.
    Functions under the assumption data is hard assigned by a group classifer with errors
    and the alignment need not perfectly correspond to groups
    parameters
    ----------
    y_true: a numpy array containing the target labels
    proba: a numpy array containing the soft classifier responses.
    metrics: an iterable of BaseGroupMetrics
    hard_assignment: a potentially lossy assignment of datapoints to groups by a classifier.
    true_groups: a numpy array containing the actual group assignment
    group_response: (optional) The response used by a classifier to soft assign groups.
    steps: (optional) The number of divisions per group
    directions: (optional) a binary vector containing [+1,-1] indicating if greater or lower
        solutions are prefered
    """
    assert proba.shape[1] == 2
    assert proba.ndim == 2
    assert y_true.ndim == 1
    assert y_true.shape[0] == proba.shape[0]
    points = y_true.shape[0]
    assert hard_assignment.shape[0] == points
    assert hard_assignment.ndim == 1
    assert true_groups.shape[0] == points
    assert true_groups.ndim == 1
    score = proba[:, 0] - proba[:, 1]
    if group_response is not False:
        assert group_response.ndim == 1
        assert points == group_response.shape[0]
        score /= group_response  # generally not useful

    unweighted_path = all([m.cond_weights is None for m in metrics])

    # hard assignment and true groups need to be ints
    # All supported metrics are invariant to the peturbation of true groups
    # ordering. So we do not need to assume that they arrive as ints, we just encode
    # and discard the ordering later.

    encoder = OrdinalEncoder()
    encoder.fit(true_groups.reshape(-1, 1))
    true_groups = encoder.transform(true_groups.reshape(-1, 1)).reshape(-1).astype(int)
    assigned_labels = np.arange(hard_assignment.max()+1)
    groups = true_groups.max() + 1
    uniq = np.unique(hard_assignment)
    if uniq.size < assigned_labels.size:
        logger.warning('Some groups were not assigned, we only saw: %s', np.array2string(uniq))

    if groups > assigned_labels.size:
        logger.warning("Fewer groups used (%d) in infered groups than in the true groups (%d)",
                       assigned_labels.size, groups)
    elif groups + 1 < assigned_labels.size:
        logger.warning("Substantially fewer groups (%d) used in true groups than in the infered groups (%d)",
                       groups, assigned_labels.size)

    ass_size = assigned_labels.shape[0]

    # Preamble that reorganizes the data for efficient computation
    # This uses lists indexed by group rather than arrays
    # as there are different amounts of data per group

    masks = [hard_assignment == g for g in assigned_labels]
    # We need to use assigned_labels rather than uniq for consistency with the offsets
    # used in fair.py.

    if unweighted_path:
        collate = [condense(score[m], y_true[m], 2, true_groups[m], groups) for m in masks]
    else:
        assert factor is not None, 'Called fit with conditional metrics but no conditional factor provided'
        # Consider disabling this and just use weight=1 if no factor provided
        weights = [1,] * len(metrics)
        for i, met in enumerate(metrics):
            if met.cond_weights is not None:
                weights[i] = met.cond_weights(factor, true_groups, y_true)

        def mask_weight(weight, mask):
            if isinstance(weight, np.ndarray):
                return weight[mask]
            else:
                return weight
        collate = [condense_weights(score[m], y_true[m], 2, true_groups[m], groups,
                                    weights=[mask_weight(w, m) for w in weights]) for m in masks]
    thresholds = [c[0] for c in collate]
    ordered_encode2: Union[bool, List]
    if unweighted_path:
        ordered_encode = [c[1] for c in collate]
        ordered_encode2 = False
    else:
        ordered_encode = [c[1][0] for c in collate]
        ordered_encode2 = [c[1][1] for c in collate]

    thresholds = [np.concatenate((t[0:1] + 1e-4, t), 0) for t in thresholds]
    # add threshold above maximum value

    if unweighted_path:
        score, indicies, front, index = grid_search_no_weights(ordered_encode, ass_size, score,
                                                               metrics, steps, directions, additional_constraints)
    else:
        score, indicies, front, index = grid_search_weights(ordered_encode, ordered_encode2,
                                                            ass_size, score, metrics,
                                                            steps, directions, additional_constraints)

    new_front, new_index = keep_front(score, indicies, directions, additional_constraints)
    # merge the two existing fronts
    front, index = keep_front(np.concatenate((front, new_front), 1),
                              np.concatenate((index, new_index), 1),
                              directions, additional_constraints)

    def av_thresh(thresh, index):
        if thresh.size == 0:
            return np.zeros(index.shape)
        iplus1 = np.minimum(thresh.shape[0] - 1, index + 1)
        return (thresh[index] + thresh[iplus1]) / 2

    selected_thresholds = np.asarray([av_thresh(g, i)
                                      for g, i in zip(thresholds, index)])
    return front, selected_thresholds



"""The entry point to fair. Defines the FairPredictor object used to access fairness
functionality."""
from ast import Tuple
from typing import Optional
import logging
import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder
from ..utils import group_metrics
from .. utils.scipy_metrics_cont_wrapper import ScorerRequiresContPred
from ..utils.group_metric_classes import BaseGroupMetric

from ..utils import performance as perf
from . import efficient_compute, fair_frontier

AUTOGLUON_EXISTS = True
try:
    from autogluon.core.metrics import Scorer
    from autogluon.tabular import TabularPredictor
except ModuleNotFoundError:
    AUTOGLUON_EXISTS = False

logger = logging.getLogger(__name__)


class FairPredictor:
    """Assess and mitigate the unfairness and effectiveness of a binary predictor
    post-fit by computing group specific metrics, and performing threshold adjustment.
    Parameters
    ----------
    predictor: a binary  predictor that will be evaluated and modified. This can be:
        1. An autogluon binary predictor.
        2. A sklearn classifier.
        3. An arbitary function
        4. The value None.
        If None  is used, we assume that we are rescoring predictions already made elsewhere, and
        the validation data should be a copy of the classifier outputs.

    validation_data: This can be:
        1. a pandas dataframe that can be read by predictor.
        2. a dict contain mutliple entries
            'data' containing a pandas dataframe or numpy array to be fed to the classifier.
            'target' the ground truth-labels used to evaluate classifier peformance.
            'groups' (optional)
            'cond_factor'
    groups (optional, default None): is an indicator of protected attributes, i.e.  the discrete
        groups used to measure fairness
    it may be:
        1. The name of a pandas column containing discrete values
        2. a vector of the same size as the validation set containing discrete values
        3. The value None   (used when we don't require groups, for example,
          if we are optimizing F1 without per-group thresholds, or if groups are explicitly
               specified by a dict in validation data)
    inferred_groups: (Optional, default False) A binary or multiclass autogluon predictor that
                    infers the protected attributes.
        This can be used to enforce fairness when no information about protected attribtutes is
        avalible at test time. If this is not false, fairness will be measured using the variable
        'groups', but enforced using the predictor response.
    use_fast: (Optional, default True) Bool or 'hybrid'
        If use_fast is True, the fair search is much more efficient, but the objectives must take the
        form of a GroupMetric. 'Hybrid' initialises the slow_pathway with the output of the fast pathhway
        This is useful for infered groups in deep networks.
        If use_fast is False, autogluon and scikitlearn scorers are also supported.
    conditioning_factor (optional, default None) Used to specify the factor conditional metrics
        are conditioned on.
        Takes the same form as groups.
    threshold (optional, default 2/3) used in use_fast pathway. Adds an extra catagory of uncertain
        group when infering attributes.
        If a datapoint has no response from the inferred_groups classifier above the threshold
        then it is assigned to the uncertain group. Tuning this value may help improve
        fairness/performance trade-offs.
        When set to 0 it is off.
    """

    def __init__(self, predictor, validation_data, groups=None, *, inferred_groups=False,
                 add_noise=False, logit_scaling=False,
                 use_fast=True, conditioning_factor=None, threshold=2/3) -> None:
        if predictor is None:
            def predictor(x):
                return x
        if not is_not_autogluon(predictor) and predictor.problem_type != 'binary':
            logger.error('Fairpredictor only takes a binary predictor as input')

        assert use_fast == 'hybrid' or use_fast is False or use_fast is True
        # Check if sklearn
        _guard_predictor_data_match(validation_data, predictor)
        self.predictor = predictor
        if groups is None:
            if isinstance(validation_data, dict):
                if isinstance(validation_data.get('groups', None), str):
                    groups = validation_data['groups']
            else:
                groups = False
        else:
            if isinstance(validation_data, dict) and validation_data.get('groups', None) is not None:
                logger.warning(("Groups passed twice to fairpredictor both as part of "
                                "the dataset and as an argument. "
                                "The argument will be used."))
        # Internal logic differentiates between groups should be recovered from other data
        # i.e. groups = None
        # and there are no groups i.e. groups = False
        # However, as a user interface groups = None makes more sense for instantiation.
        # This colides with the alternative use case where a dict is passed and groups
        # are estimated.
        if use_fast == 'hybrid':
            self.threshold = 0
        else:
            self.threshold = threshold

        self.groups = groups
        self.use_fast: bool = use_fast
        self.conditioning_factor = conditioning_factor
        self.logit_scaling = logit_scaling
        if isinstance(validation_data, dict):
            self.validation_data = validation_data  # ['data']
            validation_labels = validation_data['target']
            if groups is None:
                groups = validation_data.get('groups', False)
                # Do not update self.groups otherwise this will stick
            if conditioning_factor is None:
                conditioning_factor = validation_data.get('cond_fact', False)
                # Do not update self.conditioning otherwise this will stick
            else:
                if validation_data.get('cond_fact', False) is not False:
                    logger.warning(("Conditioning factor passed twice to fairpredictor both as "
                                    "part of the dataset and as an argument. "
                                    "The argument will be used."))
        else:
            self.validation_data = validation_data
            validation_labels = self.validation_data[predictor.label]

        # We use _internal_groups as a standardized argument that is always safe to pass
        # to functions expecting a vector
        self._internal_groups = self.groups_to_numpy(groups, self.validation_data)
        self._internal_conditioning_factor = self.cond_fact_to_numpy(conditioning_factor, self.validation_data)
        if self._internal_groups.shape[0] != validation_labels.shape[0]:
            logger.error('The size of the groups does not match the dataset size')

        self.inferred_groups = inferred_groups
        if inferred_groups:
            self._val_thresholds = call_or_get_proba(inferred_groups, self.validation_data)
        else:
            # Use OneHot and store encoder so it will work on new data
            self.group_encoder: OneHotEncoder = OneHotEncoder(handle_unknown='ignore')
            self.group_encoder.fit(self._internal_groups.reshape(-1, 1))
            self._val_thresholds = self.group_encoder.transform(
                self._internal_groups.reshape(-1, 1)).toarray()
        self.proba = call_or_get_proba(predictor, self.validation_data)
        self.add_noise = add_noise
        if add_noise:
            self.proba += np.random.normal(0, add_noise, self.proba.shape)
        if is_not_autogluon(self.predictor):
            self.y_true = np.asarray(validation_labels)
        else:
            self.y_true = np.asarray(validation_labels == self.predictor.class_labels[1])
        self.frontier: Optional[Tuple] = None
        if self.use_fast is True:
            self.offset = np.zeros((self._val_thresholds.shape[1],))
        else:
            self.offset = np.zeros((self._val_thresholds.shape[1], self.proba.shape[1]))
        self.objective1 = None
        self.objective2 = None
        self.round = False

    def _to_numpy(self, x, data, name='groups', none_replace=None) -> Optional[np.ndarray]:
        """helper function for transforming groups into a numpy array of unique values
        parameters
        ----------
        x: a standard represenations such as might be used for groups (see class doc)
        data: a pandas dataframe or a dict containing data
        returns
        -------
        numpy array
        """
        if data is None:
            data = self.validation_data
        if x is None and isinstance(none_replace, str):
            x = none_replace
        if x is None and isinstance(data, dict):
            x = data.get(name, None)
        if x is None:
            x = none_replace

        if isinstance(data, dict):
            data = data['data']
        if x is False:
            return np.zeros(data.shape[0])

        if callable(x):
            return self.infered_to_hard(x(data))
        if isinstance(x, str):
            return np.asarray(data[x])
        if isinstance(x, int):
            return np.asarray(data[:, x])
        if x is None:
            return None
        return np.asarray(x)

    def groups_to_numpy(self, groups, data):
        """helper function for transforming groups into a numpy array of unique values
        parameters
        ----------
        groups: one of the standard represenations of groups (see class doc)
        data: a pandas dataframe, numpy array, or a dict containing data
        returns
        -------
        numpy array
        """
        return self._to_numpy(groups, data, 'groups', self.groups)

    def cond_fact_to_numpy(self, fact, data):
        """helper function for transforming fact into a numpy array of unique values
        parameters
        ----------
        fact: one of the standard represenations of conditioning factor
        data: a pandas dataframe, numpy array, or a dict containing data
        returns
        -------
        numpy array
        """
        out = self._to_numpy(fact, data, 'cond_fact', self.conditioning_factor)
        # conditioning factor may be arbitary unique values like groups,
        if out is None:
            return out
        from sklearn.preprocessing import OrdinalEncoder
        encoder = OrdinalEncoder()
        out = encoder.fit_transform(out.reshape(-1, 1))
        return out.reshape(-1)

    def infered_to_hard(self, infered):
        "Map the output of infered groups into a hard assignment for use in the fast pathway"
        if self.inferred_groups is False or self.threshold == 0:
            return infered.argmax(1)

        drop = infered.max(1) < self.threshold
        out = infered.argmax(1)+1
        out[drop] = 0
        return out

    def fit(self, objective, constraint=group_metrics.accuracy, value=0.0, *,
            greater_is_better_obj=None, greater_is_better_const=None,
            recompute=True, tol=False, grid_width=False, threshold=None,
            additional_constraints=()):
        """Fits the chosen predictor to optimize an objective while satisfing a constraint.
        parameters
        ----------
        objective: a BaseGroupMetric or Scorable to be optimised
        constraint (optional): a BaseGroupMetric or Scorable that must be above/below a certain
        value
        value (optional): float the value constraint must be above or below
        If neither constraint nor value are provided fit enforces the constraint that accuracy
        is greater or equal to zero.

        greater_is_better_obj: bool or None Governs if the objective is maximised (True) or
                             minimized (False).
                If None the value of objective.greater_is_better is used.
        greater_is_better_const: bool or None Governs if the constraint has to be greater (True) or
                                smaller (False) than value.
                If None the value of constraint.greater_is_better is used.
        recompute: governs if the the parato frontier should be recomputed. Use False to efficiently
                    adjusting the threshold while keeping objective and constraint fixed.
        tol: float or False. Can round the solutions found by predict_proba to within a particular
                            tolerance to prevent overfitting.
                               Generally not needed.
        grid_width: allows manual specification of the grid size. N.B. the overall computational
                    budget is O(grid_width**groups)
                 By default the grid_size is 30
        threshold: A float between 0 and 1 or None. If threshold is not None, this overwrites
                    the threshold used for assignment to a "don't know class" in the hard assignment
                    of inferened groups.
        returns
        -------
        Nothing
        """
        if threshold is not None:
            self.threshold = threshold
        if greater_is_better_obj is None:
            greater_is_better_obj = objective.greater_is_better
        if greater_is_better_const is None:
            greater_is_better_const = constraint.greater_is_better

        if recompute is True or self.frontier is None:
            self.compute_frontier(objective, constraint,
                                  greater_is_better_obj1=greater_is_better_obj,
                                  greater_is_better_obj2=greater_is_better_const, tol=tol,
                                  grid_width=grid_width,
                                  additional_constraints=additional_constraints)
        if greater_is_better_const:
            mask = self.frontier[0][1] >= value
        else:
            mask = self.frontier[0][1] <= value

        if mask.sum() == 0:
            logger.warning("No solutions satisfy the constraint found, selecting the" +
                           " closest solution.")
            weights = self.frontier[1]
            vmax = [self.frontier[0][1].argmin(),
                    self.frontier[0][1].argmax()][int(greater_is_better_const)]
        else:
            values = self.frontier[0][0][mask]
            weights = self.frontier[1].T[mask].T

            vmax = [values.argmin(),
                    values.argmax()][int(greater_is_better_obj)]
        self.offset = weights.T[vmax].T

    def compute_frontier(self, objective1, objective2, greater_is_better_obj1,
                         greater_is_better_obj2, *, tol=False,
                         grid_width=False, additional_constraints=()) -> None:
        """ Computes the parato frontier. Internal logic used by fit
        parameters
        ----------
        objective1: a BaseGroupMetric or Scorable to be optimised
        objective2: a BaseGroupMetric or Scorable to be optimised
        greater_is_better_obj1: bool or None Governs if the objective is maximised (True)
                                 or  minimized (False).
                If None the value of objective.greater_is_better is used.
        greater_is_better_obj2: bool or None Governs if the constraint has to be greater (True)
                                or  smaller (False) than value.
                If None the value of constraint.greater_is_better is used.
        tol: float or False. Can round the solutions found by predict_proba to within a given
                            tolerance to prevent overfitting
                            Generally not needed.
        grid_width: allows manual specification of the grid size. N.B. the overall computational
                    budget is O(grid_width**groups)
        returns
        -------
        Nothing
        """
        self.objective1 = objective1
        self.objective2 = objective2
        objectives = (objective1, objective2) + tuple((a[0] for a in additional_constraints))

        direction = np.ones(2 + len(additional_constraints))
        values = np.ones(len(additional_constraints))
        if greater_is_better_obj1 is False:
            direction[0] = -1
        if greater_is_better_obj2 is False:
            direction[1] = -1

        for i, c in enumerate(additional_constraints):
            assert 3 >= len(c) >= 2
            if len(c) == 2:
                if c[0].greater_is_better is False:
                    direction[2+i] = -1
            else:
                assert c[2] == '<' or c[2] == '>'
                if c[2] == '<':
                    direction[2+i] = -1
            values[i] = c[1]

        proba = self.proba
        self.round = tol

        if tol is not False:
            proba = np.around(self.proba / tol) * tol

        def call_slow(existing_weights=None):
            fix_obj = [fix_groups_and_conditioning(obj, self._internal_groups,
                                                   self._internal_conditioning_factor, self.y_true) for obj in objectives]
            if grid_width is False:
                gw = 18
            else:
                gw = grid_width
            coarse_thresh = np.asarray(self._val_thresholds, dtype=np.float16)
            return fair_frontier.build_coarse_to_fine_front(fix_obj,
                                                            self.y_true, proba,
                                                            coarse_thresh,
                                                            directions=direction,
                                                            nr_of_recursive_calls=3,
                                                            initial_divisions=gw,
                                                            logit_scaling=self.logit_scaling,
                                                            existing_weights=existing_weights,
                                                            additional_constraints=values)

        def call_fast(grid_width=grid_width):
            if grid_width is False:
                grid_width = min(30, (30**5)**(1 / self._val_thresholds.shape[1]))
            return efficient_compute.grid_search(self.y_true, proba, objectives,
                                                 self.infered_to_hard(self._val_thresholds),
                                                 self._internal_groups,
                                                 steps=min(30, (30**5)**(1 / self._val_thresholds.shape[1])),
                                                 directions=direction, factor=self._internal_conditioning_factor,
                                                 additional_constraints=values)

        if self.use_fast == 'hybrid':
            frontier = call_fast(min(grid_width, min(30, (30**5)**(1 / self._val_thresholds.shape[1]))))
            weights = frontier[1]
            new_weights = np.zeros((weights.shape[0], 2, weights.shape[1]), dtype=weights.dtype)
            new_weights[::-1, 0, :] = weights
            self.frontier = call_slow(new_weights)
        elif self.use_fast is True:
            self.frontier = call_fast()
        else:
            self.frontier = call_slow()

    def plot_frontier(self, data=None, groups=None, *, objective1=False, objective2=False,
                      show_updated=True, show_original=True, color=None, new_plot=True, prefix='',
                      name_frontier='Frontier', subfig=None) -> None:
        """ Plots an existing parato frontier with respect to objective1 and objective2.
            These do not need to be the same objectives as used when computing the frontier
            The original predictor, and the predictor selected by fit is shown in different colors.
            fit() must be called first.
            parameters
            ----------
            data: (optional) pandas dataset or dict. If not specified, uses the data used to run fit.
            groups: (optional) groups data (see class definition). If not specified, uses the
                                definition provided at initialisation
            objective1: (optional) an objective to be plotted, if not specified use the
                                    objective provided to fit is used in its place.
            objective2: (optional) an objective to be plotted, if not specified use the
                                    constraint provided to fit is used in its place.
            show_updated: (optional, default True) Highlight the updated classifier with a
                different marker
            color: (optional, default None) Specify the color the frontier should be plotted in.
            new_plot: (optional, default True) specifies if plt.figure() should be called at the
            start or if an existing plot should be overlayed
            prefix (optional string) an additional prefix string that will be added to the legend
            for frontier and updated predictor.
        """
        import matplotlib.pyplot as plt  # noqa: C0415
        assert self.frontier is not None, 'Call fit before plot_frontier.'
        _guard_predictor_data_match(data, self.predictor)
        objective1 = objective1 or self.objective1
        objective2 = objective2 or self.objective2
        if not subfig and new_plot:
            plt.figure()
        if subfig:
            ax = subfig
            ax.set_title('Frontier found')
            ax.set_ylabel(objective1.name)
            ax.set_xlabel(objective2.name)
        else:
            ax = plt
            ax.title('Frontier Found')
            ax.ylabel(objective1.name)
            ax.xlabel(objective2.name)

        if data is None:
            data = self.validation_data
            labels = self.y_true
            proba = self.proba
            groups = self.groups_to_numpy(groups, data)
            if groups is None:
                groups = np.ones_like(labels)
            val_thresholds = self._val_thresholds
        else:
            if isinstance(data, dict):
                labels = np.asarray(data['target'])
                proba = call_or_get_proba(self.predictor, data['data'])

            else:
                labels = np.asarray(data[self.predictor.label])
                proba = call_or_get_proba(self.predictor, data)
                labels = (labels == self.predictor.positive_class) * 1
            if self.add_noise:
                proba += np.random.normal(0, self.add_noise, proba.shape)

            groups = self.groups_to_numpy(groups, data)
            if groups is None:
                groups = np.ones_like(labels)

            if self.inferred_groups is False:
                if self.groups is False:
                    val_thresholds = np.ones((groups.shape[0], 1))
                else:
                    val_thresholds = self.group_encoder.transform(groups.reshape(-1, 1)).toarray()
            else:
                if isinstance(data, dict):
                    val_thresholds = call_or_get_proba(self.inferred_groups, data['data'])
                else:
                    val_thresholds = call_or_get_proba(self.inferred_groups, data)
        if self.use_fast is not True:
            factor = self._internal_conditioning_factor
            if _needs_groups(objective1):
                objective1 = fix_groups_and_conditioning(objective1,
                                                         self.groups_to_numpy(groups, data), factor, self.y_true)
            if _needs_groups(objective2):
                objective2 = fix_groups_and_conditioning(objective2,
                                                         self.groups_to_numpy(groups, data), factor, self.y_true)

            front1 = fair_frontier.compute_metric(objective1, labels, proba,
                                                  val_thresholds, self.frontier[1])
            front2 = fair_frontier.compute_metric(objective2, labels, proba,
                                                  val_thresholds, self.frontier[1])

            zero = [dispatch_metric(objective1, labels, proba, groups, factor),
                    dispatch_metric(objective2, labels, proba, groups, factor)]

            front1_u = fair_frontier.compute_metric(objective1, labels, proba,
                                                    val_thresholds, self.offset[:, :, np.newaxis])
            front2_u = fair_frontier.compute_metric(objective2, labels, proba,
                                                    val_thresholds, self.offset[:, :, np.newaxis])

        else:
            front1 = efficient_compute.compute_metric(objective1, labels, proba,
                                                      groups,
                                                      self.infered_to_hard(val_thresholds),
                                                      self.frontier[1])
            front2 = efficient_compute.compute_metric(objective2, labels, proba,
                                                      groups,
                                                      self.infered_to_hard(val_thresholds),
                                                      self.frontier[1])

            zero = [objective1(labels, proba.argmax(1), groups),
                    objective2(labels, proba.argmax(1), groups)]

            front1_u = efficient_compute.compute_metric(objective1, labels, proba, groups,
                                                        self.infered_to_hard(val_thresholds),
                                                        self.offset[:, np.newaxis])
            front2_u = efficient_compute.compute_metric(objective2, labels, proba, groups,
                                                        self.infered_to_hard(val_thresholds),
                                                        self.offset[:, np.newaxis])
        if color is None:
            ax.scatter(front2, front1, label=prefix+name_frontier)
            if show_original:
                ax.scatter(zero[1], zero[0], s=40, label='Original predictor', marker='*')
            if show_updated:
                ax.scatter(front2_u, front1_u, s=40, label=prefix+'Updated predictor', marker='s')
            ax.legend(loc='best')
        else:
            ax.scatter(front2, front1, c=color)

    def evaluate(self, data=None, metrics=None, verbose=True) -> pd.DataFrame:
        """Compute standard metrics of the original predictor and the updated predictor
         found by fit and return them in a dataframe.
          If fit has not been called only return the metrics of the original predictor.
        parameters
        ----------
        data: (optional) a pandas dataframe to evaluate over. If not provided evaluate over
            the dataset provided at initialisation.
        metrics: (optional) a dictionary where the keys are metric names and the elements are either
                    scoreables or group metrics. If not provided report the standard metrics
                    reported by autogluon on binary predictors
        returns
        -------
        a pandas dataset containing rows indexed by metric name, and columns by
        ['original', 'updated']
         """
        _guard_predictor_data_match(data, self.predictor)
        if metrics is None:
            metrics = group_metrics.ag_metrics
        return self.evaluate_fairness(data, metrics=metrics, verbose=verbose)

    def evaluate_fairness(self, data=None, groups=None, factor=None, *,
                          metrics=None, verbose=True) -> pd.DataFrame:
        """Compute standard fairness metrics for the orginal predictor and the new predictor
         found by fit. If fit has not been called return a dataframe containing
         only the metrics of the original predictor.
         parameters
        ----------
        data: (optional) a pandas dataframe to evaluate over. If not provided evaluate over
                the dataset provided at initialisation.
        groups (optional) a specification of the groups (see class defintion). If not provided use
                the defintion provided at init.
        metrics: (optional) a dictionary where the keys are metric names and the elements are either
                    scoreables or group metrics. If not provided report the standard metrics
                    reported by SageMaker Clarify
                    https://mkai.org/learn-how-amazon-sagemaker-clarify-helps-detect-bias
        returns
        -------
        a pandas dataset containing rows indexed by fairness measure name, and columns by
        ['original', 'updated']
         """
        _guard_predictor_data_match(data, self.predictor)
        factor = self.cond_fact_to_numpy(factor, data)
        if metrics is None:
            metrics = group_metrics.default_fairness_measures

        if data is None:
            data = self.validation_data
            labels = self.y_true
            y_pred_proba = call_or_get_proba(self.predictor, data)
        else:
            if isinstance(data, dict):
                labels = data['target']
                y_pred_proba = call_or_get_proba(self.predictor, data['data'])
            else:
                labels = np.asarray(data[self.predictor.label])
                y_pred_proba = call_or_get_proba(self.predictor, data)
                if not is_not_autogluon(self.predictor):
                    labels = (labels == self.predictor.positive_class) * 1
        groups = self.groups_to_numpy(groups, data)
        score = y_pred_proba[:, 1] - y_pred_proba[:, 0]
        collect = perf.evaluate_fairness(labels, score, groups, factor,
                                         metrics=metrics, verbose=verbose, threshold=0)
        collect.columns = ['original']

        if self.frontier is not None:
            y_pred_proba = np.asarray(self.predict_proba(data))
            score = y_pred_proba[:, 1]-y_pred_proba[:, 0]
            new_pd = perf.evaluate_fairness(labels, score, groups, factor,
                                            metrics=metrics, verbose=verbose,
                                            threshold=0)

            new_pd.columns = ['updated']
            collect = pd.concat([collect, new_pd], axis='columns')
        return collect

    def fairness_metrics(self, y_true: np.ndarray, proba, groups: np.ndarray,
                         metrics, factor, *, verbose=True) -> pd.DataFrame:
        """Helper function for evaluate_fairness
        Report fairness metrics that do not require additional information.
        parameters
        ----------
        y_true: numpy array containing true binary labels of the dataset
        proba: numpy or pandas array containing the output of predict_proba
        groups: numpy array containing discrete group labelling
        metrics: a dictionary where keys are the names and values are either
        Scorable or a BaseGroupMetric.
        returns
        -------
        a pandas dataframe of fairness metrics
        """
        values = np.zeros(len(metrics))
        names = []
        for i, k in enumerate(metrics.keys()):
            if verbose is False:
                names.append(k)
            else:
                names.append(metrics[k].name)
            values[i] = dispatch_metric(metrics[k], y_true, proba, groups, factor)

        return pd.DataFrame(values, index=names)

    def evaluate_groups(self, data=None, groups=None, metrics=None, fact=None, *,
                        return_original=True, verbose=True):
        """Evaluate standard metrics per group and returns dataframe.
        parameters
        ----------
        data: (optional) a pandas dataframe to evaluate over. If not provided evaluate over
            the dataset provided at initialisation.
        groups (optional) a specification of the groups (see class defintion). If not provided
                use the defintion provided at init.
        metrics: (optional) a dictionary where the keys are metric names and the elements are either
                    scoreables or group metrics. If not provided report the standard autogluon
                    binary predictor evaluations plus measures of the size of each group and their
                    labels.
        return_original: (optional) bool.
                            If return_original is true, it returns a hierarchical dataframe
                            of the scores of the original classifier under key 'original'and the
                            scores of the updated classifier under key 'updated'.
                            If return_original is false it returns a dataframe of the scores of the
                            updated classifier only.
        returns
        -------
        either a dict of pandas dataframes or a single pandas dataframe, depending on the value of
        return original.
        """
        _guard_predictor_data_match(data, self.predictor)
        if metrics is None:
            metrics = group_metrics.default_group_metrics
        if data is None:
            data = self.validation_data
            y_true = self.y_true
            new_pred_proba = np.asarray(self.predict_proba(data))
            if return_original:
                orig_pred_proba = np.asarray(call_or_get_proba(self.predictor, data))
        else:
            if isinstance(data, dict):
                y_true = data['target']
                new_pred_proba = np.asarray(self.predict_proba(data))
                if return_original:
                    orig_pred_proba = call_or_get_proba(self.predictor, data['data'])
            else:
                y_true = np.asarray(data[self.predictor.label])
                new_pred_proba = np.asarray(self.predict_proba(data))
                if return_original:
                    orig_pred_proba = np.asarray(call_or_get_proba(self.predictor, data))
                y_true = (y_true == self.predictor.positive_class) * 1

        if self.add_noise and return_original:
            orig_pred_proba += np.random.normal(0, self.add_noise, orig_pred_proba.shape)

        groups = self.groups_to_numpy(groups, data)
        fact = self.cond_fact_to_numpy(fact, data)
        if groups is None:
            groups = np.ones_like(y_true, dtype=int)
        if return_original:
            score = orig_pred_proba[:, 1] - orig_pred_proba[:, 0]
            original = perf.evaluate_per_group(y_true, score, groups,
                                               fact,
                                               threshold=0,
                                               metrics=metrics,
                                               verbose=verbose)

        score = new_pred_proba[:, 1] - new_pred_proba[:, 0]
        updated = perf.evaluate_per_group(y_true, score, groups,
                                          fact,
                                          threshold=0,
                                          metrics=metrics,
                                          verbose=verbose)

        out = updated
        if return_original:
            out = pd.concat([original, updated], keys=['original', 'updated'])
        return out

    def predict_proba(self, data, *, transform_features=True, force_normalization=False):
        """Duplicates the functionality of predictor.predict_proba for fairpredictor.
        parameters
        ----------
        data a pandas array to make predictions over.
        return
        ------
        a  pandas array of scores. Note, these scores are not probabilities, and not guarenteed to
        be non-negative or to sum to 1.
        To make them positive and sum to 1 use force_normalization=True
        """
        if self.groups is None and self.inferred_groups is False:
            _guard_predictor_data_match(data, self.predictor)
        if self.groups is None and isinstance(data, dict):
            groups = data.get('groups', False)
        else:
            groups = self.groups
        if isinstance(data, dict):
            data = data['data']

        if is_not_autogluon(self.predictor):
            proba = call_or_get_proba(self.predictor, data)
        else:
            proba: pd.DataFrame = self.predictor.predict_proba(data,
                                                               transform_features=transform_features)
        if self.add_noise:
            proba += np.random.normal(0, self.add_noise, proba.shape)

        if self.inferred_groups is False:
            if groups is False:
                onehot = np.ones((data.shape[0], 1))
            else:
                groups = self.groups_to_numpy(groups, data)
                onehot = self.group_encoder.transform(groups.reshape(-1, 1)).toarray()
        else:
            if isinstance(data, dict):
                onehot = call_or_get_proba(self.inferred_groups, data['data'])
            else:
                onehot = call_or_get_proba(self.inferred_groups, data)
        if self.use_fast is True:
            tmp = np.zeros_like(proba)
            cache = self.offset[self.infered_to_hard(onehot)]
            if force_normalization:
                tmp[:, 1] = np.maximum(cache, 0)
                tmp[:, 0] = np.maximum(-cache, 0)
            else:
                tmp[:, 1] = cache
        else:
            tmp2 = onehot.dot(self.offset)
            if force_normalization:
                tmp = np.zeros_like(proba)
                tmp[:, 1] = np.maximum(tmp2[:, 1] - tmp2[:, 0], 0)
                tmp[:, 0] = np.maximum(tmp2[:, 0] - tmp2[:, 1], 0)
            else:
                tmp = tmp2
        if self.round is not False:
            proba = np.around(proba / self.round) * self.round
        proba += tmp
        if force_normalization:
            sum = proba.sum(1)
            if isinstance(proba, pd.DataFrame):
                proba[proba.columns[0]] /= sum
                proba[proba.columns[1]] /= sum
            else:
                proba /= sum[:, np.newaxis]

        return proba

    def predict(self, data, *, transform_features=True) -> pd.Series:
        "duplicates the functionality of predictor.predict for fairpredictor"
        proba = self.predict_proba(data, transform_features=transform_features)
        if isinstance(proba, pd.DataFrame):
            return proba.idxmax(1)
        return np.argmax(proba, 1)

    def extract_coefficients(self):
        """Extracts coefficients used to combine the heads when creating a fair deep classifier.
        This code assumes only two groups and that second head of the model is trained to output single
        values with target values 0 and 1 corresponding to membership of one of two protected groups.
        If instead the second head returns  a 1-hot encoding, indicating membership of 2 or more groups,
        use extract_coefficients_1_hot.
        This code does not support objects created with use_fast=True.
        Returns two coefficients.
        1. a scalar a, and
        2. bias term b.
        Such that head_1 + a * head_2 + b has the same outputs as our fair classifier.
        This can be used to merge the coefficients of the two heads, creating a single-headed fair classifier.
        """
        return self.offset[1, 0]-self.offset[0, 0], -self.offset[1, 0]

    def extract_coefficients_1_hot(self):
        """Extracts coefficients used to combine the heads when creating a fair deep classifier.
        This code assumes that second head of the model is trained to output a one hot encoding
        corresponding to membership of a protected group.
        For more compact binary encodings see extract_coeefficents
        This code does not support objects created with use_fast=True.
        Returns a vector coefficient a.
        Such that head_1 + a.dot(head_2) has the same outputs as our fair classifier.
        This can be used to merge the coefficients of the two heads, creating a single-headed fair classifier.
        """
        return self.offset[:, 0]


def _needs_groups(func) -> bool:
    """Internal helper function. Check if a metric is a scorer. If not assume it requires a group
    argument.
    parameters
    ----------
    func either a Scorable or GroupMetric
    """
    if not AUTOGLUON_EXISTS:
        return True
    return not isinstance(func, Scorer)


def is_not_autogluon(predictor) -> bool:
    """Internal helper function. Checks if a predictor is not an autogluon fuction."""
    if AUTOGLUON_EXISTS:
        return not isinstance(predictor, TabularPredictor)
    return True


def call_or_get_proba(predictor, data) -> np.ndarray:
    """Internal helper function. Implicit dispatch depending on if predictor is callable
    or follows scikit-learn interface.
    Converts output to numpy array"""
    if isinstance(data, dict):
        data = data['data']
    if callable(predictor):
        return np.asarray(predictor(data))
    return np.asarray(predictor.predict_proba(data))


def _guard_predictor_data_match(data, predictor) -> None:
    if (data is not None
        and is_not_autogluon(predictor)
        and not (isinstance(data, dict) and
                 data.get('data', False) is not False and
                 data.get('target', False) is not False)):
        logger.error("""When not using autogluon data must be a dict containing keys
                        'data' and 'target'""")
        assert False


def inferred_attribute_builder(train, target, protected, *args, **kwargs):
    """Helper function that trains tabular predictors suitible for use when the protected attribute
        is inferred when enforcing fairness.
        parameters
        ----------
        train: a pandas dataframe
        target: a string identifying the column of the dataframe the predictor should try to
        estimate.
        protected: a string identifying the column of the dataframe that represents the
        protected attribute.
        returns
        -------
        a pair of autogluon tabular predictors.
            1. a predictor predicting the target that doesn't use the protected attribute
            2. a predictor predicting the protected attribute that doesn't use the target.

        """
    assert AUTOGLUON_EXISTS, 'Builder only works if autogluon is installed'
    target_train = train.drop(protected, axis=1, inplace=False)
    protected_train = train.drop(target, axis=1, inplace=False)
    target_predictor = TabularPredictor(label=target).fit(train_data=target_train, *args, **kwargs)
    protected_predictor = TabularPredictor(label=protected)
    protected_predictor.fit(train_data=protected_train, *args, **kwargs)
    return target_predictor, protected_predictor


def groups_to_masks(groups):
    "helper function to convert a sequence of groups to a 1-hot encoded set of masks"
    groups = np.asarray(groups)
    unique = np.unique(groups)
    mask = np.zeros((unique.shape[0], groups.shape[0]))
    for i, group_name in enumerate(unique):
        mask[i] = groups == group_name
    return mask.T


def fix_groups(metric, groups):
    """fixes the choice of groups so that BaseGroupMetrics can be passed as Scorable analogs to the
    slow pathway.

    Parameters
    ----------
    metric: a BaseGroupMetric
    groups: a 1D pandas dataframe or numpy array

    Returns
    -------
    a function that takes y_true and y_pred as an input.

        todo: return scorable"""
    if (isinstance(metric, ScorerRequiresContPred) or
       (AUTOGLUON_EXISTS and isinstance(metric, Scorer) and (metric.needs_pred is False))):
        return metric

    groups = groups_to_masks(groups)

    def new_metric(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        return metric(y_true, y_pred, groups)
    return new_metric


def fix_conditioning(metric: BaseGroupMetric, conditioning_factor):
    """fixes the choice of groups so that BaseGroupMetrics can be passed as Scorable analogs to the
    slow pathway.

    Parameters
    ----------
    metric: a BaseGroupMetric
    groups: a 1D pandas dataframe or numpy array

    Returns
    -------
    a function that takes y_true and y_pred as an input.

        todo: return scorable"""
    if (isinstance(metric, ScorerRequiresContPred) or
       (AUTOGLUON_EXISTS and isinstance(metric, Scorer)) or metric.cond_weights is None):
        logger.warning("Fixing conditoning factor on a metric that doesn't use it.")
        return metric
    conditioning_factor = np.asarray(conditioning_factor)

    def new_metric(y_true: np.ndarray, y_pred: np.ndarray, groups) -> np.ndarray:
        return metric(y_true, y_pred, groups, conditioning_factor)
    return new_metric


def fix_groups_and_conditioning(metric, groups, conditioning_factor, y_true):
    """fixes the choice of groups and conditioning factor so that BaseGroupMetrics can be passed as
    Scorable analogs to the slow pathway.

    Parameters
    ----------
    metric: a BaseGroupMetric
    groups: a 1D pandas dataframe or numpy array

    Returns
    -------
    a function that takes y_true and y_pred as an input.

        todo: return scorable"""
    if (isinstance(metric, ScorerRequiresContPred) or
       (AUTOGLUON_EXISTS and isinstance(metric, Scorer)) or metric.cond_weights is None):
        return fix_groups(metric, groups)

    conditioning_factor = np.asarray(conditioning_factor)
    groups = np.asarray(groups)
    weights = metric.cond_weights(conditioning_factor, groups, y_true)
    groups = groups_to_masks(groups)

    def new_metric(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
        return metric(y_true, y_pred, groups, weights)
    return new_metric


def dispatch_metric(metric, y_true, proba, groups, factor) -> float:
    """Helper function for making sure different types of Scorer and GroupMetrics get the right data

    Parameters
    ----------
    metric: a BaseGroupMetric or Scorable
    y_true: a binary numpy array indicating positive or negative labels
    proba: a 2xdatapoints numpy or pandas array
    groups: a numpy array indicating group membership.

    Returns
    -------
     a numpy array containing the score provided by metrics
    """
    proba = np.asarray(proba)
    try:
        if isinstance(metric, BaseGroupMetric):
            if metric.cond_weights is None:
                return metric(y_true, proba.argmax(1), groups)[0]
            return metric(y_true, proba.argmax(1), groups, factor)[0]

        if (AUTOGLUON_EXISTS and (isinstance(metric, Scorer) and (metric.needs_pred is False)) or
           isinstance(metric, group_metrics.ScorerRequiresContPred)):
            return metric(y_true, proba[:, 1] - proba[:, 0])

        return metric(y_true, proba.argmax(1))
    except ValueError:
        return np.nan


def single_threshold(x) -> np.ndarray:
    """A helper function. Allows you to measure and enforces fairness and performance measures
    by altering a single threshold for all groups.
    To use call FairPredictor with the argument infered_groups=single_threshold"""
    return np.ones((x.shape[0], 1))


def DataDict(target, data, groups=None, conditioning_factor=None) -> dict:
    "Helper function that builds dictionaries for use with sklearn classifiers"
    assert target.shape[0] == data.shape[0]
    assert data.ndim == 2
    assert target.ndim == 1 or (target.ndim == 2 and target.shape[1] == 1)
    target = np.asarray(target).reshape(-1)
    out = {'target': target, 'data': data}
    if groups is not None:
        if not isinstance(groups, str):
            assert target.shape[0] == groups.shape[0]
            assert groups.ndim == 1 or (groups.ndim == 2 and groups.shape[1] == 1)
            out['groups'] = np.asarray(groups).reshape(-1)
        else:
            out['groups'] = groups
    if conditioning_factor is not None:
        if not isinstance(conditioning_factor, str):
            assert conditioning_factor.ndim == 1
            assert target.shape[0] == conditioning_factor.shape[0]
        out['cond_fact'] = conditioning_factor
    return out


def DeepDataDict(target, score, groups, groups_inferred=None, *,
                 conditioning_factor=None) -> dict:
    """Wrapper around build_data_dict for deeplearning with inferred attributes.
     It transforms the input data into a dict, and creates helper functions so
     fairpredictor treats them appropriately.
     target: a numpy array containing the values the classifier should predict(AKA groundtruth)
     score: a numpy array that is either size n by 1, and contains a logit output or n by (1 + #groups)
     and is a concatination of the logit output with the inferered groups.
     groups: a numpy array containing true group membership.
     infered_groups: optional numpy array of size n by #groups. If score is n by 1, infered groups go here.
    """
    assert score.ndim == 2
    assert target.ndim == 1
    assert groups.ndim == 1
    assert score.shape[0] == target.shape[0]
    assert target.shape[0] == groups.shape[0]
    if groups_inferred is not None:
        assert score.shape[1] == 1
        assert groups_inferred.ndim == 2
        assert target.shape[0] == groups_inferred.shape[0]
        data = np.stack((score, groups_inferred), 1)
    else:
        # assert score.shape[1] > 1, 'When groups_inferred is None, score must also contain inferred group information'
        data = score
    return DataDict(target, data, groups, conditioning_factor=conditioning_factor)


def DeepFairPredictor(target, score, groups, groups_inferred=None,
                      *, conditioning_factor=None, truncate_logits=15,
                      use_actual_groups=False, use_fast=None,
                      logit_scaling=False) -> FairPredictor:
    """Wrapper around FairPredictor for deeplearning with inferred attributes.
     It transforms the input data into a dict, and creates helper functions so
     fairpredictor treats them appropriately.
     target: a numpy array containing the values the classifier should predict(AKA groundtruth)
     score: a numpy array that is either size n by 1, and contains a logit output or n by (1 + #groups)
            and is a concatination of the logit output with the inferered groups.
     groups: a numpy array containing true group membership.
     infered_groups: optional numpy array of size n by #groups. If score is n by 1, infered groups go here.
     truncated_logits: for performance reasons we truncate the logits to lie in [-15,15] by default. Change this here.
     use_actual_groups: bool or 'single_threshold'. Indicates if we should use actual groups, inferred groups,
                or a single global threshold for all datapoints, to enforce fairness.
     use_fast: True, False or 'hybrid' (hybrid is prefered for infered groups. Initialises the slow pathway
            with the output of the fast pathway). By default 'hybrid' unless use_actual_groups is true, in which
            case True
     """
    val_data = DeepDataDict(target, score, groups, groups_inferred, conditioning_factor=conditioning_factor)

    def square_align(array):
        return np.stack((array[:, 1], 1-array[:, 1]), 1)

    def mult_group(array):
        return array[:, 1:]

    if groups_inferred:
        if groups_inferred.shape[1] == 1:
            group_fn = square_align
        else:
            group_fn = mult_group
    else:
        if score.shape[1] == 2:
            group_fn = square_align
        else:
            group_fn = mult_group

    def capped_identity(array):
        array = np.minimum(array[:, 0], truncate_logits)
        array = np.maximum(array, -truncate_logits)
        return np.stack((-array / 2, array / 2), 1)

    if use_fast is None:
        if use_actual_groups:
            use_fast = True
        else:
            use_fast = 'hybrid'
    if use_actual_groups is True:
        fpred = FairPredictor(capped_identity, val_data, threshold=0, use_fast=use_fast, logit_scaling=logit_scaling)
    elif use_actual_groups == 'single_threshold':
        fpred = FairPredictor(capped_identity, val_data, inferred_groups=single_threshold,
                              threshold=0, use_fast=use_fast, logit_scaling=False)
    else:
        fpred = FairPredictor(capped_identity, val_data, inferred_groups=group_fn, threshold=0,
                              use_fast=use_fast, logit_scaling=logit_scaling)
    return fpred



"""Definitions of conditional measures for fairness and performance"""

from .group_metric_classes import (
    ConditionalWeighting,
)  # pylint: disable=unused-import # noqa
from . import group_metrics as gm


def reweight_by_factor_size(group_positives, group_negatives,
                            intersectional_positives, intersectional_negatives):
    """Used to rescore notions that depend on total number of entries, e.g. Positive Decision Rate"""
    return ((group_positives + group_negatives) /
            max(1, intersectional_positives + intersectional_negatives))


def reweight_by_factor_postives(group_positives, group_negatives,
                                intersectional_positives, intersectional_negatives):
    """Used to rescore notions that depend on total number of positive entries e.g. recall"""
    return (intersectional_positives) / max(1, group_positives)


def reweight_by_factor_negatives(group_positives, group_negatives,
                                 intersectional_positives, intersectional_negatives):
    """Used to rescore notions that depend on total number of negative entries e.g. sensitivity"""
    return (intersectional_negatives) / max(1, group_negatives)


def reweight_by_1(group_positives, group_negatives,
                  intersectional_positives, intersectional_negatives):
    """Used for debugging"""
    return 1


total_weights = ConditionalWeighting(reweight_by_factor_size)
pos_weights = ConditionalWeighting(reweight_by_factor_postives)
neg_weights = ConditionalWeighting(reweight_by_factor_negatives)
constant = ConditionalWeighting(reweight_by_1)


def build_cond_form(metric, weighting):
    "Build a conditional form from an existing metric using a given weighting"
    name = "Conditional " + metric.name
    return metric.clone(name, weighting)


pos_pred_rate = build_cond_form(gm.pos_pred_rate, total_weights)
neg_pred_rate = build_cond_form(gm.neg_pred_rate, total_weights)
pos_data_rate = build_cond_form(gm.pos_data_rate, total_weights)
neg_data_rate = build_cond_form(gm.neg_data_rate, total_weights)

true_pos_rate = build_cond_form(gm.true_pos_rate, pos_weights)
true_neg_rate = build_cond_form(gm.true_neg_rate, neg_weights)
false_pos_rate = build_cond_form(gm.false_pos_rate, neg_weights)
false_neg_rate = build_cond_form(gm.false_neg_rate, neg_weights)


accuracy = build_cond_form(gm.accuracy, total_weights)

cond_measures = {
    "accuracy": accuracy,
    "true_pos_rate": true_pos_rate,
    "true_neg_rate": true_neg_rate,
    "false_pos_rate": false_pos_rate,
    "false_neg_rate": false_neg_rate,
    "pos_data_rate": pos_data_rate,
    "pos_pred_rate": pos_pred_rate,
}


cond_disparities = dict(
    (key + "_diff", item.diff) for (key, item) in cond_measures.items()
)

pos_data_proportion = build_cond_form(gm.pos_data_proportion, total_weights)
neg_data_proportion = build_cond_form(gm.neg_data_proportion, total_weights)
diff_data_proportion = build_cond_form(gm.diff_data_proportion, total_weights)
abs_diff_data_proportion = build_cond_form(gm.abs_diff_data_proportion, total_weights)
pos_pred_proportion = build_cond_form(gm.pos_pred_proportion, total_weights)
neg_pred_proportion = build_cond_form(gm.neg_pred_proportion, total_weights)
diff_pred_proportion = build_cond_form(gm.diff_pred_proportion, total_weights)
abs_diff_pred_proportion = build_cond_form(gm.abs_diff_pred_proportion, total_weights)

wachter_measures = {'pos_data_proportion': pos_data_proportion,
                    'neg_data_proportion': neg_data_proportion,
                    'diff_data_proportion': diff_data_proportion,
                    'pos_pred_proportion': pos_pred_proportion,
                    'neg_pred_proportion': neg_pred_proportion,
                    'diff_pred_proportion': diff_pred_proportion}



"""Slow pathway for computing fairness constraints. Compatable with Scorers and group metrics,
while efficient_compute is only compatable with group metrics"""
from typing import Callable, Tuple, Sequence, Optional, Union
import numpy as np
from .efficient_compute import keep_front
from ..utils.scipy_metrics_cont_wrapper import ScorerRequiresContPred
AUTOGLUON_EXISTS = True
try:
    from autogluon.core.metrics import Scorer
except ModuleNotFoundError:
    AUTOGLUON_EXISTS = False


def compute_metric(metric: Callable, y_true: np.ndarray, proba: np.ndarray,
                   threshold_assignment: np.ndarray,
                   weights: np.ndarray) -> np.ndarray:
    """takes probability scores, and offsets them according to the weights * threshold_assignment.
        then select the max and compute a fairness metric """
    score = np.zeros((weights.shape[-1]))
    y_true = np.asarray(y_true)
    threshold_assignment = np.asarray(threshold_assignment)

    pass_scores = (isinstance(metric, ScorerRequiresContPred) or
                   (AUTOGLUON_EXISTS and isinstance(metric, Scorer) and (metric.needs_pred is False)))
    # Preallocate because this next loop is the system bottleneck
    tmp = np.empty((threshold_assignment.shape[0], weights.shape[1]), dtype=threshold_assignment.dtype)
    pred = np.empty(threshold_assignment.shape[0], dtype=int)
    for i in range(weights.shape[-1]):
        np.dot(threshold_assignment, weights[:, :, i], tmp)
        if pass_scores is False:
            tmp += proba
            np.argmax(tmp, -1, pred)
            score[i] = metric(y_true, pred)[0]
        else:
            tmp += proba
            tmp[:, 1] -= tmp[:, 0]
            score[i] = metric(y_true, tmp[:, 1])
    return score


def compute_metrics(metrics: Sequence[Callable], y_true: np.ndarray, proba: np.ndarray,
                    threshold_assignment: np.ndarray,
                    weights: np.ndarray) -> np.ndarray:
    """takes probability scores, and offsets them according to the weights * threshold_assignment.
        then select the max and compute a fairness metric """
    scores = np.zeros((len(metrics), weights.shape[-1]))
    y_true = np.asarray(y_true)
    assert weights[:, 1, :].sum() == 0
    weights = weights[:, 0, :]
    proba = proba[:, 0] - proba[:, 1]

    threshold_assignment = np.asarray(threshold_assignment)

    pass_scores = [(isinstance(metric, ScorerRequiresContPred) or
                   (AUTOGLUON_EXISTS and isinstance(metric, Scorer) and (metric.needs_pred is False)))
                   for metric in metrics]
    # Preallocate because this next loop is the system bottleneck
    tmp = np.empty((threshold_assignment.shape[0],), dtype=threshold_assignment.dtype)
    # diff = np.empty(threshold_assignment.shape[0], dtype=threshold_assignment.dtype)
    # pred = np.empty(threshold_assignment.shape[0], dtype=int)
    for i in range(weights.shape[-1]):
        threshold_assignment.dot(weights[:, i], out=tmp)
        tmp += proba  # [:, np.newaxis]
        pred = (tmp <= 0)

        # np.dot(threshold_assignment, weights[:, i], tmp)
        # tmp += proba
        for j, metric in enumerate(metrics):
            if pass_scores[j] is False:
                scores[j, i] = metric(y_true, pred)[0]
            else:
                # np.subtract(tmp[:, 1], tmp[:, 0], diff)
                scores[j, i] = metric(y_true, tmp)
    return scores


def sort_by_front(front: np.ndarray, weights: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    "sort the front and weights according to front[0]"
    sort_ind = np.argsort(front[0])
    weights = weights[:, :, sort_ind]
    front = front[:, sort_ind]
    return front, weights

# Solution modified from here:
# https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python


def linear_interpolate(front: np.ndarray, weights: np.ndarray, gap: float = 0.01) -> np.ndarray:
    """we want the points found to cover the frontier i.e. there should be no big gaps in w.
        we achieve this by linearly interpolating in the frontier, and using this to determine
        step size in w """
    eps = (front.max(1) - front.min(1)) * gap  # step size in the frontier
    out = [np.linspace(weights[:, :, i], weights[:, :, i + 1],
                       num=int((np.abs(front[:, i + 1] - front[:, i]) / eps).max()) + 1)
           for i in range(weights.shape[-1] - 1)]
    out_np = np.concatenate(out, 0).transpose(1, 2, 0)
    return out_np


def sigmoid(x: np.ndarray) -> np.ndarray:
    "broadcastable sigmoid function"
    return 1.0 / (1.0 + np.exp(-x))


def inv_sigmoid(x) -> np.ndarray:
    "broadcastable inverse sigmoid function"
    assert (x < 1).all()
    assert (x > 0).all()
    return np.log(x / (1 - x))


def make_grid_between_points(point_a: np.ndarray, point_b: np.ndarray, refinement_factor: Union[int, np.ndarray], *,
                             add_zero: bool = False, use_linspace: bool = True,
                             logit_scaling: bool = False) -> np.ndarray:
    """
    creates new grid points between two points by refining each axis in which the two points do not
    coincide; the refinement_factor defines into how many parts such an axis is divided
    """
    assert point_a.shape == point_b.shape
    groups, classes = point_a.shape
    point_a = point_a.flatten()
    point_b = point_b .flatten()
    mins = np.minimum(point_a, point_b)
    maxs = np.maximum(point_a, point_b)
    if use_linspace:
        axx = make_grid_linspace(mins, maxs, logit_scaling, refinement_factor, add_zero)
    else:
        axx = make_grid_arange(mins, maxs, refinement_factor, add_zero)

    mesh = np.meshgrid(*axx, copy=False)
    shape = (groups, classes + 1) + mesh[0].shape
    weights = np.zeros(shape, dtype=np.float16)
    for i in range(classes):  # Ignore final class -- the space of thresholds is overparameterized
        for j in range(groups):
            weights[j, i] = mesh[(classes) * j + i]
    weights = weights.reshape((weights.shape[0], weights.shape[1], -1))
    assert not np.isnan(weights).any()
    return weights


def make_grid_linspace(mins: np.ndarray, maxs: np.ndarray, logit_scaling: bool,
                       refinement_factor: Union[int, np.ndarray], add_zero: bool) -> Sequence[np.ndarray]:
    if logit_scaling:
        maxs = sigmoid(maxs)
        mins = sigmoid(mins)
    diffs = maxs - mins
    if logit_scaling:
        epsilon = 0.0
    else:
        epsilon = 0.005
        if any(diffs > 0):
            epsilon = diffs[diffs > 0].min()
    mins -= epsilon
    maxs += epsilon
    zero = np.zeros((1))
    if logit_scaling:
        zero = 0.5 * np.ones(1)
        axx = [inv_sigmoid(np.linspace(mins[i], maxs[i], num=refinement_factor + 1))
               for i in range(maxs.shape[0])]
    else:
        axx = [(np.linspace(mins[i], maxs[i], num=refinement_factor + 1))
               for i in range(maxs.shape[0])]
    if add_zero:
        axx = [np.concatenate((ax, zero), 0) for ax in axx]
    return axx


def make_grid_arange(mins: np.ndarray, maxs: np.ndarray, refinement_factor: np.ndarray,
                     add_zero: bool) -> Sequence[np.ndarray]:
    epsilon = refinement_factor.flatten()
    mins -= epsilon * 1.5
    maxs += epsilon * 1.51
    mins -= epsilon
    maxs += epsilon
    zero = np.zeros((1))
    axx = [np.arange(mins[i], maxs[i], step=epsilon[i]) for i in range(maxs.shape[0])]
    if add_zero:
        axx = [np.concatenate((ax, zero), 0) for ax in axx]
    return axx


def make_finer_grid(weights: np.ndarray, refinement_factor: Union[int, np.ndarray] = 2,
                    use_linspace: bool = True) -> np.ndarray:
    """
    creates additional grid points between two consecutive points in the current weights set; see
    the function make_grid_between_points below for the meaning of the refinement_factor
    """
    new_weights = [make_grid_between_points(weights[:, :-1, ell],
                                            weights[:, :-1, ell + 1],
                                            refinement_factor=refinement_factor,
                                            use_linspace=use_linspace)
                   for ell in range(weights.shape[-1] - 1)]
    output = np.concatenate(new_weights, axis=2)
    output = np.concatenate((output, np.zeros((output.shape[0], output.shape[1], 1), dtype=np.float16)), axis=2)
    output = np.unique(output, axis=-1)

    return output


def front_from_weights(weights: np.ndarray, y_true: np.ndarray, proba: np.ndarray,
                       groups_infered: np.ndarray,
                       tupple_metrics: Sequence[Callable]) -> np.ndarray:
    """Computes the values of each metric from the weights"""
    front = compute_metrics(tupple_metrics, y_true, proba, groups_infered, weights)
    # front = np.stack(list(map(lambda x: compute_metric(x, y_true, proba,
    #                                                   groups_infered, weights), tupple_metrics)))
    return front


def build_coarse_to_fine_front(metrics: Sequence[Callable],
                               y_true: np.ndarray,
                               proba: np.ndarray,
                               groups_infered: np.ndarray,
                               directions: np.ndarray,
                               *,
                               initial_divisions: int = 15,
                               nr_of_recursive_calls: int = 5,
                               refinement_factor: int = 4,
                               logit_scaling: bool = False,
                               existing_weights: Optional[np.ndarray] = None,
                               additional_constraints: Sequence = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    this function performs coarse-to-fine grid-search for computing the Pareto front
    """
    assert len(metrics) >= 2
    assert groups_infered.ndim == 2
    assert nr_of_recursive_calls > 0
    groups = groups_infered.shape[1]

    classes = proba.shape[1] - 1  # n.b. this is really classes-1
    upper_bound = (proba[:, :classes] - proba.min(1)[:, np.newaxis]).max(0)
    lower_bound = (proba[:, :classes] - proba.max(1)[:, np.newaxis]).min(0)
    min_initial = np.ones((groups, classes))
    min_initial[:, :] = lower_bound[:, np.newaxis]
    max_initial = np.ones((groups, classes))
    max_initial[:, :] = upper_bound[:, np.newaxis]
    # perform an initial two stage search, first coarsely over every possible value
    # then take the front and search over valid values from it
    weights = make_grid_between_points(min_initial, max_initial,
                                       refinement_factor=initial_divisions - 1,
                                       logit_scaling=logit_scaling)
    front = front_from_weights(weights, y_true, proba, groups_infered, metrics)
    front, weights = keep_front(front, weights, directions, additional_constraints)
    # second stage
    mins = weights[:, :-1].min(-1)  # drop zeros
    maxs = weights[:, :-1].max(-1)
    mins -= 2 / initial_divisions  # if we only get one point, expand around it
    maxs += 2 / initial_divisions
    eps = ((maxs - mins))
    new_weights = make_grid_between_points(mins, maxs, refinement_factor=initial_divisions,
                                           add_zero=True, logit_scaling=logit_scaling)
    new_front = front_from_weights(new_weights, y_true, proba, groups_infered, metrics)
    weights = np.concatenate((new_weights, weights), -1)
    front = np.concatenate((new_front, front), -1)
    if existing_weights is not None:
        existing_weights = existing_weights.astype(weights.dtype)
        existing_front = front_from_weights(existing_weights, y_true, proba, groups_infered, metrics)
        weights = np.concatenate((existing_weights, weights), -1)
        front = np.concatenate((existing_front, front), -1)

    front, weights = keep_front(front, weights, directions, additional_constraints)
    for _ in range(nr_of_recursive_calls - 1):
        if weights.shape[-1] != 1:
            eps /= refinement_factor
            new_weights = make_finer_grid(weights, eps, use_linspace=False)
            new_front = front_from_weights(new_weights, y_true, proba, groups_infered,
                                           metrics)
            weights = np.concatenate((new_weights, weights), -1)
            front = np.concatenate((new_front, front), -1)
            front, weights = keep_front(front, weights, directions, additional_constraints)

    # densify the front with uniform interpolation
    if weights.shape[-1] > 1:
        weights = linear_interpolate(front, weights, gap=0.02)
        front = front_from_weights(weights, y_true, proba, groups_infered, metrics)
        front, weights = keep_front(front, weights, directions, additional_constraints)

    return front, weights



import os

from setuptools import setup

FAIR = "oxonfair"

version = "0.2.1"

PYTHON_REQUIRES = ">=3.8"


def create_version_file(*, version):
    print("-- Building version " + version)
    version_path = os.path.join("src", FAIR, "version.py")
    with open(version_path, "w") as f:
        f.write(f'"""This is the {FAIR} version file."""\n')
        f.write("__version__ = '{}'\n".format(version))


def update_version(version, use_file_if_exists=True, create_file=False):
    """
    To release a new stable version on PyPi, simply tag the release on github, and the Github CI will automatically publish
    a new stable version to PyPi using the configurations in .github/workflows/pypi_release.yml. **Currently disabled**.
    You need to increase the version number after stable release, so that the nightly pypi can work properly.
    """
    try:
        if not os.getenv("RELEASE"):
            from datetime import date

            minor_version_file_path = "VERSION.minor"
            if use_file_if_exists and os.path.isfile(minor_version_file_path):
                with open(minor_version_file_path) as f:
                    day = f.read().strip()
            else:
                today = date.today()
                day = today.strftime("b%Y%m%d")
            version += day
    except Exception:
        pass
    if create_file and not os.getenv("RELEASE"):
        with open("VERSION.minor", "w") as f:
            f.write(day)
    return version


def default_setup_args(*, version):
    from setuptools import find_packages

    long_description = open("README.md").read()
    name = f"{FAIR}"
    setup_args = dict(
        name=name,
        version=version,
        author="Governance of Emerging Technologies Programme (Oxford Internet Insitute)",
        url="https://github.com/oxfordinternetinstitute/oxonfair",
        description="Toolkit for evaluating and enforcing ML model fairness",
        long_description=long_description,
        long_description_content_type="text/markdown",
        license="Apache-2.0",
        license_files=("LICENSE", "NOTICE"),
        # Package info
        package_dir={"": "src"},
        packages=find_packages("src"),
        namespace_packages=[],
        zip_safe=True,
        include_package_data=True,
        python_requires=PYTHON_REQUIRES,
        package_data={
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Education",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Customer Service",
            "Intended Audience :: Financial and Insurance Industry",
            "Intended Audience :: Healthcare Industry",
            "Intended Audience :: Telecommunications Industry",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Scientific/Engineering :: Image Recognition",
        ],
        project_urls={
            "Documentation": "https://github.com/oxfordinternetinstitute/oxonfair",
            "Bug Reports": "https://github.com/oxfordinternetinstitute/oxonfair/issues",
            "Source": "https://github.com/oxfordinternetinstitute/oxonfair",
            "Contribute!": "https://github.com/oxfordinternetinstitute/oxonfair/blob/master/CONTRIBUTING.md",
        },
    )
    return setup_args


install_requires = [
    "numpy>=1.21.4",
    "pandas>=1.2.5",
    "scikit-learn",
    'ucimlrepo'
]

extras_require = dict()
full_requirements = ['matplotlib', 'autogluon.tabular', 'torch', 'xgboost', 'jupyterlab']
notebook_requirements = full_requirements + ['fairlearn']
test_requirements = notebook_requirements + ["tox", "pytest", "pytest-cov", 'flake8',
                                             'linkcheckmd', 'ipynbcompress',]

full_requirements = list(set(full_requirements))
notebook_requirements = list(set(notebook_requirements))
test_requirements = list(set(test_requirements))
extras_require['full'] = full_requirements
extras_require['notebooks'] = notebook_requirements
extras_require["tests"] = test_requirements


if __name__ == "__main__":
    create_version_file(version=version)
    setup_args = default_setup_args(version=version)
    setup(
        install_requires=install_requires,
        extras_require=extras_require,
        **setup_args,
    )



from .fair import (FairPredictor, inferred_attribute_builder, single_threshold, DataDict,
                   DeepFairPredictor, DeepDataDict)

__all__ = (FairPredictor, inferred_attribute_builder, single_threshold, DataDict, DeepFairPredictor,
           DeepDataDict)
